# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utility function for WCS

"""

import functools
import re
import warnings

import astropy.units as u
import numpy as np
from astropy import coordinates as coords
from astropy.io import fits
from astropy.modeling import core, projections
from astropy.modeling import models as astmodels
from astropy.wcs import Celprm

# these ctype values do not include yzLN and yzLT pairs
sky_pairs = {
    "equatorial": ["RA", "DEC"],
    "ecliptic": ["ELON", "ELAT"],
    "galactic": ["GLON", "GLAT"],
    "helioecliptic": ["HLON", "HLAT"],
    "supergalactic": ["SLON", "SLAT"],
    # "spec": specsystems
}

radesys = ["ICRS", "FK5", "FK4", "FK4-NO-E", "GAPPT", "GALACTIC"]


class UnsupportedTransformError(Exception):
    def __init__(self, message):
        super().__init__(message)


class UnsupportedProjectionError(Exception):
    def __init__(self, code):
        message = f"Unsupported projection: {code}"
        super().__init__(message)


class RegionError(Exception):
    def __init__(self, message):
        super().__init__(message)


class CoordinateFrameError(Exception):
    def __init__(self, message):
        super().__init__(message)


def to_index(value):
    """
    Convert value to an int or an int array.

    Input coordinates converted to integers
    corresponding to the center of the pixel.
    The convention is that the center of the pixel is
    (0, 0), while the lower left corner is (-0.5, -0.5).
    The outputs are used to index the mask.

    Examples
    --------
    >>> to_index(np.array([-0.5, 0.49999]))
    array([0, 0])
    >>> to_index(np.array([0.5, 1.49999]))
    array([1, 1])
    >>> to_index(np.array([1.5, 2.49999]))
    array([2, 2])
    """
    return np.asarray(np.floor(np.asarray(value) + 0.5), dtype=int)


def _toindex(value):
    msg = "_toindex is deprecated, use to_index instead."
    warnings.warn(DeprecationWarning(msg), stacklevel=2)

    return to_index(value)


def get_values(units, *args):
    """
    Return the values of Quantity objects after optionally converting to units.

    Parameters
    ----------
    units : str or `~astropy.units.Unit` or None
        Units to convert to. The input values are converted to ``units``
        before the values are returned.
    args : `~astropy.units.Quantity`
        Quantity inputs.
    """
    if units is not None:
        result = [a.to_value(unit) for a, unit in zip(args, units, strict=False)]
    else:
        result = [a.value for a in args]
    return result


def _compute_lon_pole(skycoord, projection):
    """
    Compute the longitude of the celestial pole of a standard
    frame in the native frame.

    This angle then can be used as one of the Euler angles
    (the other two being skycoord) to rotate the native frame into the
    standard frame ``skycoord.frame``.

    Parameters
    ----------
    skycoord : `astropy.coordinates.SkyCoord`, or
               sequence of floats or `~astropy.units.Quantity` of length 2
        The celestial longitude and latitude of the fiducial point - typically
        right ascension and declination. These are given by the ``CRVALia``
        keywords in ``FITS``.

    projection : `astropy.modeling.projections.Projection`
        A `~astropy.modeling.projections.Projection` model instance.

    Returns
    -------
    lonpole : float or `~astropy/units.Quantity`
        Native longitude of the celestial pole in degrees.

    """
    if isinstance(skycoord, coords.SkyCoord):
        lon = skycoord.spherical.lon.value
        lat = skycoord.spherical.lat.value
        unit = u.deg
    else:
        lon, lat = skycoord
        unit = None
        if isinstance(lon, u.Quantity):
            lon = lon.to_value(u.deg)
            unit = u.deg
        if isinstance(lat, u.Quantity):
            lat = lat.to_value(u.deg)
            unit = u.deg

    if isinstance(projection, projections.Zenithal):
        lonpole = 180
    else:
        cel = Celprm()
        cel.ref = [lon, lat]
        cel.prj.code = projection.prjprm.code
        pvrange = projection.prjprm.pvrange
        if pvrange:
            i1 = pvrange // 100
            i2 = i1 + (pvrange % 100) + 1
            cel.prj.pv = i1 * [None] + list(projection.prjprm.pv[i1:i2])
        cel.set()

        lonpole = cel.ref[2]
    if unit is not None:
        lonpole = lonpole * unit

    return lonpole


def get_projcode(wcs_info):
    # CTYPE here is only the imaging CTYPE keywords
    sky_axes, _, _ = get_axes(wcs_info)
    if not sky_axes:
        return None
    projcode = wcs_info["CTYPE"][sky_axes[0]][5:8].upper()
    if projcode not in projections.projcodes:
        msg = f"Projection code {projcode}, not recognized"
        raise UnsupportedProjectionError(msg)
    return projcode


def read_wcs_from_header(header):
    """
    Extract basic FITS WCS keywords from a FITS Header.

    Parameters
    ----------
    header : astropy.io.fits.Header
        FITS Header with WCS information.

    Returns
    -------
    wcs_info : dict
        A dictionary with WCS keywords.
    """
    wcs_info = {}

    try:
        wcs_info["WCSAXES"] = header["WCSAXES"]
    except KeyError:
        p = re.compile(r"ctype[\d]*", re.IGNORECASE)
        ctypes = header["CTYPE*"]
        keys = list(ctypes.keys())
        for key in keys[::-1]:
            if p.split(key)[-1] != "":
                keys.remove(key)
        wcs_info["WCSAXES"] = len(keys)
    wcsaxes = wcs_info["WCSAXES"]
    # if not present call get_csystem
    wcs_info["RADESYS"] = header.get("RADESYS", "ICRS")
    wcs_info["VAFACTOR"] = header.get("VAFACTOR", 1)
    wcs_info["NAXIS"] = header.get("NAXIS", 0)
    wcs_info["EQUINOX"] = header.get("EQUINOX", None)
    wcs_info["EPOCH"] = header.get("EPOCH", None)
    wcs_info["DATEOBS"] = header.get("MJD-OBS", header.get("DATE-OBS", None))

    ctype = []
    cunit = []
    crpix = []
    crval = []
    cdelt = []
    for i in range(1, wcsaxes + 1):
        ctype.append(header[f"CTYPE{i}"])
        cunit.append(header.get(f"CUNIT{i}", None))
        crpix.append(header.get(f"CRPIX{i}", 0.0))
        crval.append(header.get(f"CRVAL{i}", 0.0))
        cdelt.append(header.get(f"CDELT{i}", 1.0))

    if "CD1_1" in header:
        wcs_info["has_cd"] = True
    else:
        wcs_info["has_cd"] = False
    pc = np.zeros((wcsaxes, wcsaxes))
    for i in range(1, wcsaxes + 1):
        for j in range(1, wcsaxes + 1):
            key = f"CD{i}_{j}" if wcs_info["has_cd"] else f"PC{i}_{j}"
            if key in header:
                pc[i - 1, j - 1] = header[key]
            elif i == j:
                pc[i - 1, j - 1] = 1.0
            else:
                pc[i - 1, j - 1] = 0.0
    wcs_info["CTYPE"] = ctype
    wcs_info["CUNIT"] = cunit
    wcs_info["CRPIX"] = crpix
    wcs_info["CRVAL"] = crval
    wcs_info["CDELT"] = cdelt
    wcs_info["PC"] = pc
    return wcs_info


def get_axes(header):
    """
    Matches input with spectral and sky coordinate axes.

    Parameters
    ----------
    header : astropy.io.fits.Header or dict
        FITS Header (or dict) with basic WCS information.

    Returns
    -------
    sky_inmap, spectral_inmap, unknown : lists
        indices in the input representing sky and spectral coordinates.

    """
    if isinstance(header, fits.Header):
        wcs_info = read_wcs_from_header(header)
    elif isinstance(header, dict):
        wcs_info = header
    else:
        msg = "Expected a FITS Header or a dict."
        raise TypeError(msg)

    # Split each CTYPE value at "-" and take the first part.
    # This should represent the coordinate system.
    ctype = [ax.split("-")[0].upper() for ax in wcs_info["CTYPE"]]
    sky_inmap = []
    spec_inmap = []
    unknown = []
    skysystems = np.array(list(sky_pairs.values())).flatten()
    for ax in ctype:
        ind = ctype.index(ax)
        if ax in specsystems:
            spec_inmap.append(ind)
        elif ax in skysystems:
            sky_inmap.append(ind)
        else:
            unknown.append(ind)

    if sky_inmap:
        _is_skysys_consistent(ctype, sky_inmap)

    return sky_inmap, spec_inmap, unknown


def _is_skysys_consistent(ctype, sky_inmap):
    """Determine if the sky axes in CTYPE match to form a standard celestial system."""

    for item in sky_pairs.values():
        if ctype[sky_inmap[0]] == item[0]:
            if ctype[sky_inmap[1]] != item[1]:
                msg = "Inconsistent ctype for sky coordinates {} and {}".format(*ctype)
                raise ValueError(msg)
            break
        if ctype[sky_inmap[1]] == item[0]:
            if ctype[sky_inmap[0]] != item[1]:
                msg = "Inconsistent ctype for sky coordinates {} and {}".format(*ctype)
                raise ValueError(msg)
            sky_inmap = sky_inmap[::-1]
            break


specsystems = [
    "WAVE",
    "FREQ",
    "ENER",
    "WAVEN",
    "AWAV",
    "VRAD",
    "VOPT",
    "ZOPT",
    "BETA",
    "VELO",
]

sky_systems_map = {
    "ICRS": coords.ICRS,
    "FK5": coords.FK5,
    "FK4": coords.FK4,
    "FK4NOE": coords.FK4NoETerms,
    "GAL": coords.Galactic,
    "HOR": coords.AltAz,
}


def make_fitswcs_transform(header):
    """
    Create a basic FITS WCS transform.
    It does not include distortions.

    Parameters
    ----------
    header : astropy.io.fits.Header or dict
        FITS Header (or dict) with basic WCS information

    """
    if isinstance(header, fits.Header):
        wcs_info = read_wcs_from_header(header)
    elif isinstance(header, dict):
        wcs_info = header
    else:
        msg = "Expected a FITS Header or a dict."
        raise TypeError(msg)
    transforms = []
    wcs_linear = fitswcs_linear(wcs_info)
    transforms.append(wcs_linear)
    wcs_nonlinear = fitswcs_nonlinear(wcs_info)
    if wcs_nonlinear is not None:
        transforms.append(wcs_nonlinear)
    return functools.reduce(core._model_oper("|"), transforms)


def fitswcs_linear(header):
    """
    Create a WCS linear transform from a FITS header.

    Parameters
    ----------
    header : astropy.io.fits.Header or dict
        FITS Header or dict with basic FITS WCS keywords.

    """
    if isinstance(header, fits.Header):
        wcs_info = read_wcs_from_header(header)
    elif isinstance(header, dict):
        wcs_info = header
    else:
        msg = "Expected a FITS Header or a dict."
        raise TypeError(msg)

    pc = wcs_info["PC"]
    # get the part of the PC matrix corresponding to the imaging axes
    sky_axes, spec_axes, unknown = get_axes(wcs_info)
    if pc.shape != (2, 2):
        if sky_axes:
            i, j = sky_axes
        elif unknown and len(unknown) == 2:
            i, j = unknown
        sky_pc = np.zeros((2, 2))
        sky_pc[0, 0] = pc[i, i]
        sky_pc[0, 1] = pc[i, j]
        sky_pc[1, 0] = pc[j, i]
        sky_pc[1, 1] = pc[j, j]
        pc = sky_pc.copy()

    sky_axes.extend(unknown)
    if sky_axes:
        crpix = []
        cdelt = []
        for i in sky_axes:
            crpix.append(wcs_info["CRPIX"][i])
            cdelt.append(wcs_info["CDELT"][i])
    else:
        cdelt = wcs_info["CDELT"]
        crpix = wcs_info["CRPIX"]

    # if wcsaxes == 2:
    rotation = astmodels.AffineTransformation2D(matrix=pc, name="pc_matrix")

    translation_models = [
        astmodels.Shift(-(shift - 1), name="crpix" + str(i + 1))
        for i, shift in enumerate(crpix)
    ]
    translation = functools.reduce(lambda x, y: x & y, translation_models)

    if not wcs_info["has_cd"]:
        # Do not compute scaling since CDELT* = 1 if CD is present.
        scaling_models = [
            astmodels.Scale(scale, name="cdelt" + str(i + 1))
            for i, scale in enumerate(cdelt)
        ]

        scaling = functools.reduce(lambda x, y: x & y, scaling_models)
        wcs_linear = translation | rotation | scaling
    else:
        wcs_linear = translation | rotation

    return wcs_linear


def fitswcs_nonlinear(header):
    """
    Create a WCS linear transform from a FITS header.

    Parameters
    ----------
    header : astropy.io.fits.Header or dict
        FITS Header or dict with basic FITS WCS keywords.
    """
    if isinstance(header, fits.Header):
        wcs_info = read_wcs_from_header(header)
    elif isinstance(header, dict):
        wcs_info = header
    else:
        msg = "Expected a FITS Header or a dict."
        raise TypeError(msg)

    transforms = []
    projcode = get_projcode(wcs_info)
    if projcode is not None:
        projection = create_projection_transform(projcode).rename(projcode)
        transforms.append(projection)
    # Create the sky rotation transform
    sky_axes, _, _ = get_axes(wcs_info)
    if sky_axes:
        phip, lonp = (wcs_info["CRVAL"][i] for i in sky_axes)
        thetap = _compute_lon_pole((phip, lonp), projection)
        n2c = astmodels.RotateNative2Celestial(phip, lonp, thetap, name="crval")
        transforms.append(n2c)
    if transforms:
        return functools.reduce(core._model_oper("|"), transforms)
    return None


def create_projection_transform(projcode):
    """
    Create the non-linear projection transform.

    Parameters
    ----------
    projcode : str
        FITS WCS projection code.

    Returns
    -------
    transform : astropy.modeling.Model
        Projection transform.
    """

    projklassname = f"Pix2Sky_{projcode}"
    try:
        projklass = getattr(projections, projklassname)
    except AttributeError as err:
        raise UnsupportedProjectionError(projcode) from err

    projparams = {}
    return projklass(**projparams)


def is_high_level(*args, low_level_wcs):
    """
    Determine if args matches the high level classes as defined by
    ``low_level_wcs``.
    """
    if low_level_wcs.world_axis_object_classes is None or len(args) != len(
        low_level_wcs.world_axis_object_classes
    ):
        return False

    type_match = [
        (type(arg), waoc[0])
        for arg, waoc in zip(
            args, low_level_wcs.world_axis_object_classes.values(), strict=False
        )
    ]

    types_are_high_level = [argt is t for argt, t in type_match]

    if all(types_are_high_level):
        return True

    if any(types_are_high_level):
        msg = (
            "Invalid types were passed, got "
            f"({', '.join(tm[0].__name__ for tm in type_match)}) expected "
            f"({', '.join(tm[1].__name__ for tm in type_match)})."
        )
        raise TypeError(msg)

    return False



__all__ = [
    "GwcsBoundingBoxWarning",
    "GwcsFrameExistsError",
    "NoConvergence",
]


class NoConvergence(Exception):
    """
    An error class used to report non-convergence and/or divergence
    of numerical methods. It is used to report errors in the
    iterative solution used by
    the :py:meth:`~astropy.wcs.WCS.all_world2pix`.

    Attributes
    ----------

    best_solution : `numpy.ndarray`
        Best solution achieved by the numerical method.

    accuracy : `numpy.ndarray`
        Estimate of the accuracy of the ``best_solution``.

    niter : `int`
        Number of iterations performed by the numerical method
        to compute ``best_solution``.

    divergent : None, `numpy.ndarray`
        Indices of the points in ``best_solution`` array
        for which the solution appears to be divergent. If the
        solution does not diverge, ``divergent`` will be set to `None`.

    slow_conv : None, `numpy.ndarray`
        Indices of the solutions in ``best_solution`` array
        for which the solution failed to converge within the
        specified maximum number of iterations. If there are no
        non-converging solutions (i.e., if the required accuracy
        has been achieved for all input data points)
        then ``slow_conv`` will be set to `None`.

    """

    def __init__(
        self,
        *args,
        best_solution=None,
        accuracy=None,
        niter=None,
        divergent=None,
        slow_conv=None,
    ):
        super().__init__(*args)

        self.best_solution = best_solution
        self.accuracy = accuracy
        self.niter = niter
        self.divergent = divergent
        self.slow_conv = slow_conv


class GwcsFrameExistsError(ValueError):
    """
    An error used to report when a frame already exists in a pipeline.
    """


class GwcsBoundingBoxWarning(UserWarning):
    """
    A warning class to report issues with bounding boxes in GWCS.
    """



import warnings
from typing import NamedTuple, TypeAlias, Union

from astropy.modeling.core import Model

from gwcs.coordinate_frames import CoordinateFrame, EmptyFrame

__all__ = [
    "IndexedStep",
    "Step",
    "StepTuple",
]


StepTuple: TypeAlias = tuple[CoordinateFrame, Union[Model, None]]  # noqa: UP007


class Step:
    """
    Represents a ``step`` in the WCS pipeline.

    Parameters
    ----------
    frame : `~gwcs.coordinate_frames.CoordinateFrame`
        A gwcs coordinate frame object.
    transform : `~astropy.modeling.Model` or None
        A transform from this step's frame to next step's frame.
        The transform of the last step should be `None`.
    """

    def __init__(self, frame: str | CoordinateFrame | None, transform=None):
        # Allow for a string to be passed in for the frame but be turned into a
        # frame object
        self.frame = (
            frame if isinstance(frame, CoordinateFrame) else EmptyFrame(name=frame)
        )
        self.transform = transform

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, val):
        if not isinstance(val, CoordinateFrame | str):
            msg = '"frame" should be an instance of CoordinateFrame or a string.'
            raise TypeError(msg)

        self._frame = val

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, val):
        if val is not None and not isinstance(val, (Model)):
            msg = '"transform" should be an instance of astropy.modeling.Model.'
            raise TypeError(msg)
        self._transform = val

    @property
    def frame_name(self):
        if isinstance(self.frame, str):
            return self.frame
        return self.frame.name

    def __getitem__(self, ind):
        warnings.warn(
            "Indexing a WCS.pipeline step is deprecated. "
            "Use the `frame` and `transform` attributes instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if ind not in (0, 1):
            msg = "Allowed inices are 0 (frame) and 1 (transform)."
            raise IndexError(msg)
        if ind == 0:
            return self.frame
        return self.transform

    def __str__(self):
        return (
            f"{self.frame_name}\t "
            f"{getattr(self.transform, 'name', 'None') or type(self.transform).__name__}"  # noqa: E501
        )

    def __repr__(self):
        return (
            f"Step(frame={self.frame_name}, "
            f"transform={getattr(self.transform, 'name', 'None') or type(self.transform).__name__})"  # noqa: E501
        )

    def copy(self):
        return Step(self.frame, self.transform)


class IndexedStep(NamedTuple):
    """
    Class to handle a step and its index in the pipeline.
    """

    idx: int
    step: Step



# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else.
# See astropy.sphinx.conf for which values are set there.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('..'))
# IMPORTANT: the above commented section was generated by sphinx-quickstart, but
# is *NOT* appropriate for astropy or Astropy affiliated packages. It is left
# commented out with this explanation to make it clear why this should not be
# done. If the sys.path entry above is added, when the astropy.sphinx.conf
# import occurs, it will import the *source* version of astropy instead of the
# version installed (if invoked as "make html" or directly with sphinx), or the
# version in the build directory (if "python setup.py build_sphinx" is used).
# Thus, any C-extensions that are needed to build the documentation will *not*
# be accessible, and the documentation will not build correctly.

import sys
import tomllib
from datetime import datetime
from pathlib import Path

if sys.version_info >= (3, 12):
    from importlib.metadata import distribution
else:
    from importlib_metadata import distribution

try:
    from sphinx_astropy.conf.v2 import *  # noqa: F403
except ImportError:
    print(  # noqa: T201
        "ERROR: the documentation requires the sphinx-astropy package to be installed"
    )
    sys.exit(1)

# -- General configuration ----------------------------------------------------

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append("_templates")  # noqa: F405

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""  # noqa: F405

# Top-level directory containing ASDF schemas (relative to current directory)
asdf_schema_path = "../gwcs/schemas"
# This is the prefix common to all schema IDs in this repository
asdf_schema_standard_prefix = "stsci.edu/gwcs"
asdf_schema_reference_mappings = [
    (
        "tag:stsci.edu:asdf",
        "http://asdf-standard.readthedocs.io/en/latest/generated/stsci.edu/asdf/",
    ),
]

# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
with (Path(__file__).parent.parent / "pyproject.toml").open("rb") as metadata_file:
    configuration = tomllib.load(metadata_file)
    metadata = configuration["project"]

project = metadata["name"]
author = metadata["authors"][0]["name"]
copyright = f"{datetime.now().year}, {author}"  # noqa: A001

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

release = distribution(project).version
# for example take major/minor
version = ".".join(release.split(".")[:2])

# -- Options for HTML output ---------------------------------------------------
html_title = f"{project} v{release}"

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project}doc"

# -- Options for LaTeX output --------------------------------------------------
latex_documents = [
    ("index", project + ".tex", project + " Documentation", author, "manual")
]
latex_logo = "_static/images/logo-light-mode.png"

# -- Options for manual page output --------------------------------------------
man_pages = [("index", project.lower(), project + " Documentation", [author], 1)]

# -- Add in additional sphinx extensions ----------------------------------------
extensions += ["sphinx_asdf"]  # noqa: F405
extensions += ["sphinx_inline_tabs"]

# Enable nitpicky mode - which ensures that all references in the docs resolve.
nitpicky = True
nitpick_ignore = [
    ("py:class", "gwcs.api.GWCSAPIMixin"),
    ("py:class", "gwcs.wcs._pipeline.Pipeline"),
    ("py:obj", "astropy.modeling.projections.projcodes"),
    ("py:attr", "gwcs.WCS.bounding_box"),
    ("py:meth", "gwcs.WCS.footprint"),
]

# -- Options for html theme -------------------------------------------------
html_theme = "furo"
html_static_path = ["_static"]
# Override default settings from sphinx_asdf / sphinx_astropy (incompatible with furo)
html_sidebars = {}
# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = "_static/images/favicon.ico"
html_logo = ""

# html_theme_options = {
#     "light_logo": "images/logo-light-mode.png",
#     "dark_logo": "images/logo-dark-mode.png",
# }
html_theme_options = {
    "light_logo": "images/stsci_logo.png",
    "dark_logo": "images/stsci_logo.png",
}

pygments_style = "monokai"
# NB Dark style pygments is furo-specific at this time
pygments_dark_style = "monokai"
# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gbgcolor=white",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings
from contextlib import suppress

from asdf.extension import Converter
from asdf_astropy.converters.transform.core import (
    TransformConverterBase,
    parameter_to_value,
)

__all__ = [
    "CelestialFrameConverter",
    "CompositeFrameConverter",
    "FITSImagingWCSConverter",
    "FrameConverter",
    "SpectralFrameConverter",
    "StepConverter",
    "StokesFrameConverter",
    "TemporalFrameConverter",
    "WCSConverter",
]


class WCSConverter(Converter):
    tags = ("tag:stsci.edu:gwcs/wcs-*",)
    types = ("gwcs.wcs.WCS",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.wcs import WCS, GwcsBoundingBoxWarning

        gwcsobj = WCS(node["steps"], name=node["name"])
        if "pixel_shape" in node:
            gwcsobj.pixel_shape = node["pixel_shape"]

        # Ignore the warning about the bounding box order for data read from a
        # file. This is causing issues with files from MAST.
        with suppress(AttributeError), warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=GwcsBoundingBoxWarning)
            _ = gwcsobj.bounding_box

        return gwcsobj

    def to_yaml_tree(self, gwcsobj, tag, ctx):
        return {
            "name": gwcsobj.name,
            "steps": gwcsobj.pipeline,
            "pixel_shape": gwcsobj.pixel_shape,
        }


class StepConverter(Converter):
    tags = ("tag:stsci.edu:gwcs/step-*",)
    types = ("gwcs.wcs.Step",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.wcs import Step

        return Step(frame=node["frame"], transform=node.get("transform", None))

    def to_yaml_tree(self, step, tag, ctx):
        from gwcs.coordinate_frames import EmptyFrame

        return {
            "frame": step.frame.name
            if isinstance(step.frame, EmptyFrame)
            else step.frame,
            "transform": step.transform,
        }


class FrameConverter(Converter):
    tags = ("tag:stsci.edu:gwcs/frame-*",)
    types = ("gwcs.coordinate_frames.CoordinateFrame",)

    def _from_yaml_tree(self, node, tag, ctx):
        kwargs = {"name": node["name"]}

        if "axes_type" in node and "naxes" in node:
            kwargs.update({"axes_type": node["axes_type"], "naxes": node["naxes"]})

        if "axes_names" in node:
            kwargs["axes_names"] = node["axes_names"]

        if "reference_frame" in node:
            kwargs["reference_frame"] = node["reference_frame"]

        if "axes_order" in node:
            kwargs["axes_order"] = tuple(node["axes_order"])

        if "unit" in node:
            kwargs["unit"] = tuple(node["unit"])

        if "axis_physical_types" in node:
            kwargs["axis_physical_types"] = tuple(node["axis_physical_types"])

        return kwargs

    def _to_yaml_tree(self, frame, tag, ctx):
        from gwcs.coordinate_frames import CoordinateFrame

        node = {}

        node["name"] = frame.name

        # We want to check that it is exactly this type and not a subclass
        if type(frame) is CoordinateFrame:
            node["axes_type"] = frame.axes_type
            node["naxes"] = frame.naxes

        if frame.axes_order is not None:
            node["axes_order"] = list(frame.axes_order)

        if frame.axes_names is not None:
            node["axes_names"] = list(frame.axes_names)

        if frame.reference_frame is not None:
            node["reference_frame"] = frame.reference_frame

        if frame.unit is not None:
            node["unit"] = list(frame.unit)

        if frame.axis_physical_types is not None:
            node["axis_physical_types"] = list(frame.axis_physical_types)

        return node

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import CoordinateFrame

        node = self._from_yaml_tree(node, tag, ctx)
        return CoordinateFrame(**node)

    def to_yaml_tree(self, frame, tag, ctx):
        return self._to_yaml_tree(frame, tag, ctx)


class Frame2DConverter(FrameConverter):
    tags = ("tag:stsci.edu:gwcs/frame2d-*",)
    types = ("gwcs.coordinate_frames.Frame2D",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import Frame2D

        node = self._from_yaml_tree(node, tag, ctx)
        return Frame2D(**node)


class CelestialFrameConverter(FrameConverter):
    tags = ("tag:stsci.edu:gwcs/celestial_frame-*",)
    types = ("gwcs.coordinate_frames.CelestialFrame",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import CelestialFrame

        node = self._from_yaml_tree(node, tag, ctx)
        return CelestialFrame(**node)


class SpectralFrameConverter(FrameConverter):
    tags = ("tag:stsci.edu:gwcs/spectral_frame-*",)
    types = ("gwcs.coordinate_frames.SpectralFrame",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import SpectralFrame

        node = self._from_yaml_tree(node, tag, ctx)

        return SpectralFrame(**node)


class CompositeFrameConverter(FrameConverter):
    tags = ("tag:stsci.edu:gwcs/composite_frame-*",)
    types = ("gwcs.coordinate_frames.CompositeFrame",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import CompositeFrame

        if len(node) != 2:
            msg = "CompositeFrame has extra properties"
            raise ValueError(msg)

        name = node["name"]
        frames = node["frames"]

        return CompositeFrame(frames, name)

    def to_yaml_tree(self, frame, tag, ctx):
        return {"name": frame.name, "frames": frame.frames}


class TemporalFrameConverter(FrameConverter):
    tags = ("tag:stsci.edu:gwcs/temporal_frame-*",)
    types = ("gwcs.coordinate_frames.TemporalFrame",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import TemporalFrame

        node = self._from_yaml_tree(node, tag, ctx)
        return TemporalFrame(**node)


class StokesFrameConverter(FrameConverter):
    tags = ("tag:stsci.edu:gwcs/stokes_frame-*",)
    types = ("gwcs.coordinate_frames.StokesFrame",)

    def from_yaml_tree(self, node, tag, ctx):
        from gwcs.coordinate_frames import StokesFrame

        node = self._from_yaml_tree(node, tag, ctx)
        return StokesFrame(**node)

    def to_yaml_tree(self, frame, tag, ctx):
        node = {}

        node["name"] = frame.name
        if frame.axes_order:
            node["axes_order"] = list(frame.axes_order)

        return node


class FITSImagingWCSConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/fitswcs_imaging-*",)
    types = ("gwcs.fitswcs.FITSImagingWCSTransform",)

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.fitswcs import FITSImagingWCSTransform

        return FITSImagingWCSTransform(
            node["projection"],
            crpix=node["crpix"],
            crval=node["crval"],
            cdelt=node["cdelt"],
            pc=node["pc"],
        )

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {
            "crpix": parameter_to_value(model.crpix),
            "crval": parameter_to_value(model.crval),
            "cdelt": parameter_to_value(model.cdelt),
            "pc": parameter_to_value(model.pc),
            "projection": model.projection,
        }



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests the API defined in astropy APE 14 (https://doi.org/10.5281/zenodo.1188875).
"""

import astropy.modeling.models as m
import astropy.units as u
import numpy as np
import pytest
from astropy import coordinates as coord
from astropy import time
from astropy.wcs.wcsapi import HighLevelWCSWrapper
from astropy.wcs.wcsapi.high_level_api import values_to_high_level_objects
from numpy.testing import assert_allclose, assert_array_equal

import gwcs
import gwcs.coordinate_frames as cf


# Shorthand the name of the 2d gwcs fixture
@pytest.fixture
def wcsobj(request):
    return request.getfixturevalue(request.param)


wcs_objs = pytest.mark.parametrize("wcsobj", ["gwcs_2d_spatial_shift"], indirect=True)


@pytest.fixture
def wcs_ndim_types_units(request):
    """
    Generate a wcs and the expected ndim, types, and units.
    """
    ndim = {
        "gwcs_2d_spatial_shift": (2, 2),
        "gwcs_2d_spatial_reordered": (2, 2),
        "gwcs_1d_freq": (1, 1),
        "gwcs_3d_spatial_wave": (3, 3),
        "gwcs_4d_identity_units": (4, 4),
    }
    types = {
        "gwcs_2d_spatial_shift": ("pos.eq.ra", "pos.eq.dec"),
        "gwcs_2d_spatial_reordered": ("pos.eq.dec", "pos.eq.ra"),
        "gwcs_1d_freq": ("em.freq",),
        "gwcs_3d_spatial_wave": ("pos.eq.ra", "pos.eq.dec", "em.wl"),
        "gwcs_4d_identity_units": ("pos.eq.ra", "pos.eq.dec", "em.wl", "time"),
    }
    units = {
        "gwcs_2d_spatial_shift": ("deg", "deg"),
        "gwcs_2d_spatial_reordered": ("deg", "deg"),
        "gwcs_1d_freq": ("Hz",),
        "gwcs_3d_spatial_wave": ("deg", "deg", "m"),
        "gwcs_4d_identity_units": ("deg", "deg", "nm", "s"),
    }

    return (
        request.getfixturevalue(request.param),
        ndim[request.param],
        types[request.param],
        units[request.param],
    )


# # x, y inputs - scalar and array
x, y = 1, 2
xarr, yarr = np.ones((3, 4)), np.ones((3, 4)) + 1

fixture_names = [
    "gwcs_2d_spatial_shift",
    "gwcs_2d_spatial_reordered",
    "gwcs_1d_freq",
    "gwcs_3d_spatial_wave",
    "gwcs_4d_identity_units",
]
fixture_wcs_ndim_types_units = pytest.mark.parametrize(
    "wcs_ndim_types_units", fixture_names, indirect=True
)
all_wcses_names = [
    *fixture_names,
    "gwcs_3d_identity_units",
    "gwcs_stokes_lookup",
    "gwcs_3d_galactic_spectral",
]
fixture_all_wcses = pytest.mark.parametrize("wcsobj", all_wcses_names, indirect=True)


@fixture_all_wcses
def test_lowlevel_types(wcsobj):
    try:
        # Skip this on older versions of astropy where it doesn't exist.
        from astropy.wcs.wcsapi.tests.utils import validate_low_level_wcs_types
    except ImportError:
        return

    validate_low_level_wcs_types(wcsobj)


@fixture_all_wcses
def test_names(wcsobj):
    assert wcsobj.world_axis_names == wcsobj.output_frame.axes_names
    assert wcsobj.pixel_axis_names == wcsobj.input_frame.axes_names


def test_names_split(gwcs_3d_galactic_spectral):
    wcs = gwcs_3d_galactic_spectral
    assert (
        wcs.world_axis_names
        == wcs.output_frame.axes_names
        == ("Latitude", "Frequency", "Longitude")
    )


@fixture_wcs_ndim_types_units
def test_pixel_n_dim(wcs_ndim_types_units):
    wcsobj, ndims, *_ = wcs_ndim_types_units
    assert wcsobj.pixel_n_dim == ndims[0]


@fixture_wcs_ndim_types_units
def test_world_n_dim(wcs_ndim_types_units):
    wcsobj, ndims, *_ = wcs_ndim_types_units
    assert wcsobj.world_n_dim == ndims[1]


@fixture_wcs_ndim_types_units
def test_world_axis_physical_types(wcs_ndim_types_units):
    wcsobj, ndims, physical_types, world_units = wcs_ndim_types_units
    assert wcsobj.world_axis_physical_types == physical_types


@fixture_wcs_ndim_types_units
def test_world_axis_units(wcs_ndim_types_units):
    wcsobj, ndims, physical_types, world_units = wcs_ndim_types_units
    assert wcsobj.world_axis_units == world_units


@pytest.mark.parametrize(("x", "y"), zip((x, xarr), (y, yarr), strict=False))
def test_pixel_to_world_values(gwcs_2d_spatial_shift, x, y):
    wcsobj = gwcs_2d_spatial_shift
    assert_allclose(wcsobj.pixel_to_world_values(x, y), wcsobj(x, y))


@pytest.mark.parametrize(("x", "y"), zip((x, xarr), (y, yarr), strict=False))
def test_pixel_to_world_values_units_2d(gwcs_2d_shift_scale_quantity, x, y):
    wcsobj = gwcs_2d_shift_scale_quantity

    call_pixel = x * u.pix, y * u.pix
    api_pixel = x, y

    call_world = wcsobj(*call_pixel)
    api_world = wcsobj.pixel_to_world_values(*api_pixel)

    # Check that call returns quantities and api doesn't
    assert all(isinstance(a, u.Quantity) for a in call_world)
    assert all(not isinstance(a, u.Quantity) for a in api_world)

    # Check that they are the same (and implicitly in the same units)
    assert_allclose(u.Quantity(call_world).value, api_world)

    new_call_pixel = wcsobj.invert(*call_world)
    [assert_allclose(n, p) for n, p in zip(new_call_pixel, call_pixel, strict=False)]

    new_api_pixel = wcsobj.world_to_pixel_values(*api_world)
    [assert_allclose(n, p) for n, p in zip(new_api_pixel, api_pixel, strict=False)]


@pytest.mark.parametrize(("x"), [x, xarr])
def test_pixel_to_world_values_units_1d(gwcs_1d_freq_quantity, x):
    wcsobj = gwcs_1d_freq_quantity

    call_pixel = x * u.pix
    api_pixel = x

    call_world = wcsobj(call_pixel)
    api_world = wcsobj.pixel_to_world_values(api_pixel)

    # Check that call returns quantities and api doesn't
    assert isinstance(call_world, u.Quantity)
    assert not isinstance(api_world, u.Quantity)

    # Check that they are the same (and implicitly in the same units)
    assert_allclose(u.Quantity(call_world).value, api_world)

    new_call_pixel = wcsobj.invert(call_world)
    assert_allclose(new_call_pixel, call_pixel)

    new_api_pixel = wcsobj.world_to_pixel_values(api_world)
    assert_allclose(new_api_pixel, api_pixel)


@pytest.mark.parametrize(("x", "y"), zip((x, xarr), (y, yarr), strict=False))
def test_array_index_to_world_values(gwcs_2d_spatial_shift, x, y):
    wcsobj = gwcs_2d_spatial_shift
    assert_allclose(wcsobj.array_index_to_world_values(x, y), wcsobj(y, x))


def test_world_axis_object_components_2d(gwcs_2d_spatial_shift):
    waoc = gwcs_2d_spatial_shift.world_axis_object_components
    assert waoc[0][:2] == ("celestial", 0)
    assert callable(waoc[0][2])
    assert waoc[1][:2] == ("celestial", 1)
    assert callable(waoc[1][2])


def test_world_axis_object_components_2d_generic(gwcs_2d_quantity_shift):
    waoc = gwcs_2d_quantity_shift.world_axis_object_components
    assert waoc == [("SPATIAL", 0, "value"), ("SPATIAL1", 0, "value")]


def test_world_axis_object_components_1d(gwcs_1d_freq):
    waoc = gwcs_1d_freq.world_axis_object_components
    assert [c[:2] for c in waoc] == [("spectral", 0)]
    assert callable(waoc[0][2])


def test_world_axis_object_components_4d(gwcs_4d_identity_units):
    waoc = gwcs_4d_identity_units.world_axis_object_components
    first_two = [c[:2] for c in waoc]
    last_one = [c[2] for c in waoc]
    assert first_two == [
        ("celestial", 0),
        ("celestial", 1),
        ("spectral", 0),
        ("temporal", 0),
    ]
    assert all(callable(last) for last in last_one)


def test_world_axis_object_classes_2d(gwcs_2d_spatial_shift):
    waoc = gwcs_2d_spatial_shift.world_axis_object_classes
    assert waoc["celestial"][0] is coord.SkyCoord
    assert waoc["celestial"][1] == ()
    assert "frame" in waoc["celestial"][2]
    assert "unit" in waoc["celestial"][2]
    assert isinstance(waoc["celestial"][2]["frame"], coord.ICRS)
    assert tuple(waoc["celestial"][2]["unit"]) == (u.deg, u.deg)


def test_world_axis_object_classes_2d_generic(gwcs_2d_quantity_shift):
    waoc = gwcs_2d_quantity_shift.world_axis_object_classes
    assert waoc["SPATIAL"][0] is u.Quantity
    assert waoc["SPATIAL1"][0] is u.Quantity
    assert waoc["SPATIAL"][1] == ()
    assert waoc["SPATIAL1"][1] == ()
    assert "unit" in waoc["SPATIAL"][2]
    assert "unit" in waoc["SPATIAL1"][2]
    assert waoc["SPATIAL"][2]["unit"] == u.km
    assert waoc["SPATIAL1"][2]["unit"] == u.km


def test_world_axis_object_classes_4d(gwcs_4d_identity_units):
    waoc = gwcs_4d_identity_units.world_axis_object_classes
    assert waoc["celestial"][0] is coord.SkyCoord
    assert waoc["celestial"][1] == ()
    assert "frame" in waoc["celestial"][2]
    assert "unit" in waoc["celestial"][2]
    assert isinstance(waoc["celestial"][2]["frame"], coord.ICRS)
    assert tuple(waoc["celestial"][2]["unit"]) == (u.deg, u.deg)

    temporal = waoc["temporal"]
    assert temporal[0] is time.Time
    assert temporal[1] == ()
    assert temporal[2] == {
        "unit": u.s,
        "format": "isot",
        "scale": "utc",
        "precision": 3,
        "in_subfmt": "*",
        "out_subfmt": "*",
        "location": None,
    }


def _compare_frame_output(wc1, wc2):
    if isinstance(wc1, coord.SkyCoord):
        assert isinstance(wc1.frame, type(wc2.frame))
        assert u.allclose(wc1.spherical.lon, wc2.spherical.lon, equal_nan=True)
        assert u.allclose(wc1.spherical.lat, wc2.spherical.lat, equal_nan=True)
        assert u.allclose(
            wc1.spherical.distance, wc2.spherical.distance, equal_nan=True
        )

    elif isinstance(wc1, u.Quantity):
        assert u.allclose(wc1, wc2, equal_nan=True)

    elif isinstance(wc1, time.Time):
        assert u.allclose((wc1 - wc2).to(u.s), 0 * u.s)

    elif isinstance(wc1, str | coord.StokesCoord):
        assert wc1 == wc2

    else:
        msg = f"Can't Compare {type(wc1)}"
        raise TypeError(msg)


@fixture_all_wcses
def test_high_level_wrapper(wcsobj, request):
    hlvl = HighLevelWCSWrapper(wcsobj)

    pixel_input = [3] * wcsobj.pixel_n_dim
    if wcsobj.bounding_box is not None:
        for i, interval in wcsobj.bounding_box.intervals.items():
            bbox_min = u.Quantity(interval.lower).value
            pixel_input[i] = max(bbox_min + 1, pixel_input[i])

    # Assert that both APE 14 API and GWCS give the same answer The APE 14 API
    # uses the mixin class and __call__ calls values_to_high_level_objects
    wc1 = hlvl.pixel_to_world(*pixel_input)
    wc2 = wcsobj(*pixel_input)
    results = wcsobj._remove_units_input(wc2, wcsobj.output_frame)

    wc2 = values_to_high_level_objects(*results, low_level_wcs=wcsobj)
    if len(wc2) == 1:
        wc2 = wc2[0]
    assert type(wc1) is type(wc2)

    if isinstance(wc1, list | tuple):
        for w1, w2 in zip(wc1, wc2, strict=False):
            _compare_frame_output(w1, w2)
    else:
        _compare_frame_output(wc1, wc2)

    # we have just asserted that wc1 and wc2 are equal
    if not isinstance(wc1, list | tuple):
        wc1 = (wc1,)

    pix_out1 = hlvl.world_to_pixel(*wc1)
    np.testing.assert_allclose(pix_out1, pixel_input)
    with pytest.raises(TypeError) as e:
        _ = wcsobj.invert(*wc1)
    assert "High Level objects are not supported with the native" in str(e)


def test_stokes_wrapper(gwcs_stokes_lookup):
    hlvl = HighLevelWCSWrapper(gwcs_stokes_lookup)

    pixel_input = [0, 1, 2, 3]

    out = hlvl.pixel_to_world(pixel_input * u.pix)

    assert list(out) == ["I", "Q", "U", "V"]

    pixel_input = [
        [0, 1, 2, 3],
        [0, 1, 2, 3],
        [0, 1, 2, 3],
        [0, 1, 2, 3],
    ]

    out = hlvl.pixel_to_world(pixel_input * u.pix)

    expected = coord.StokesCoord(
        [
            ["I", "Q", "U", "V"],
            ["I", "Q", "U", "V"],
            ["I", "Q", "U", "V"],
            ["I", "Q", "U", "V"],
        ]
    )

    assert (out == expected).all()

    pixel_input = [-1, 4]

    out = hlvl.pixel_to_world(pixel_input * u.pix)

    assert np.isnan(out.value).all()

    pixel_input = [[-1, 4], [1, 2]]

    out = hlvl.pixel_to_world(pixel_input * u.pix)

    assert np.isnan(out[0].value).all()
    assert (out[1] == ["Q", "U"]).all()

    out = hlvl.pixel_to_world(1 * u.pix)

    assert isinstance(out, coord.StokesCoord)
    assert out == "Q"


@wcs_objs
def test_array_shape(wcsobj):
    assert wcsobj.array_shape is None

    wcsobj.array_shape = (2040, 1020)
    assert_array_equal(wcsobj.array_shape, (2040, 1020))

    assert wcsobj.array_shape == wcsobj.pixel_shape[::-1]

    wcsobj.pixel_shape = (1111, 2222)
    assert wcsobj.array_shape == (2222, 1111)


@wcs_objs
def test_pixel_bounds(wcsobj):
    assert wcsobj.pixel_bounds is None

    wcsobj.bounding_box = ((-0.5, 2039.5), (-0.5, 1019.5))
    assert_array_equal(wcsobj.pixel_bounds, wcsobj.bounding_box)
    # Reset the bounding box or this will affect other tests
    wcsobj.bounding_box = None


@wcs_objs
def test_axis_correlation_matrix(wcsobj):
    assert_array_equal(wcsobj.axis_correlation_matrix, np.identity(2))


@wcs_objs
def test_serialized_classes(wcsobj):
    assert not wcsobj.serialized_classes


@wcs_objs
def test_low_level_wcs(wcsobj):
    assert id(wcsobj.low_level_wcs) == id(wcsobj)


@wcs_objs
def test_pixel_to_world(wcsobj):
    values = wcsobj(x, y)
    result = wcsobj.pixel_to_world(x, y)
    assert isinstance(result, coord.SkyCoord)
    assert_allclose(values[0] * u.deg, result.data.lon)
    assert_allclose(values[1] * u.deg, result.data.lat)


@wcs_objs
def test_array_index_to_world(wcsobj):
    values = wcsobj(x, y)
    result = wcsobj.array_index_to_world(y, x)
    assert isinstance(result, coord.SkyCoord)
    assert_allclose(values[0] * u.deg, result.data.lon)
    assert_allclose(values[1] * u.deg, result.data.lat)


def test_pixel_to_world_quantity(gwcs_2d_shift_scale, gwcs_2d_shift_scale_quantity):
    result1 = gwcs_2d_shift_scale.pixel_to_world(x, y)
    result2 = gwcs_2d_shift_scale_quantity.pixel_to_world(x, y)
    assert isinstance(result2, coord.SkyCoord)
    assert_allclose(result1.data.lon, result2.data.lon)
    assert_allclose(result1.data.lat, result2.data.lat)

    # test with Quantity pixel inputs
    result1 = gwcs_2d_shift_scale.pixel_to_world(x * u.pix, y * u.pix)
    result2 = gwcs_2d_shift_scale_quantity.pixel_to_world(x * u.pix, y * u.pix)
    assert isinstance(result2, coord.SkyCoord)
    assert_allclose(result1.data.lon, result2.data.lon)
    assert_allclose(result1.data.lat, result2.data.lat)

    # test for pixel units
    with pytest.raises(ValueError):  # noqa: PT011
        gwcs_2d_shift_scale.pixel_to_world(x * u.Jy, y * u.Jy)


def test_array_index_to_world_quantity(
    gwcs_2d_shift_scale, gwcs_2d_shift_scale_quantity
):
    result0 = gwcs_2d_shift_scale.pixel_to_world(x, y)
    result1 = gwcs_2d_shift_scale.array_index_to_world(y, x)
    result2 = gwcs_2d_shift_scale_quantity.array_index_to_world(y, x)
    assert isinstance(result2, coord.SkyCoord)
    assert_allclose(result1.data.lon, result2.data.lon)
    assert_allclose(result1.data.lat, result2.data.lat)
    assert_allclose(result0.data.lon, result1.data.lon)
    assert_allclose(result0.data.lat, result1.data.lat)

    # test with Quantity pixel inputs
    result0 = gwcs_2d_shift_scale.pixel_to_world(x * u.pix, y * u.pix)
    result1 = gwcs_2d_shift_scale.array_index_to_world(y * u.pix, x * u.pix)
    result2 = gwcs_2d_shift_scale_quantity.array_index_to_world(y * u.pix, x * u.pix)
    assert isinstance(result2, coord.SkyCoord)
    assert_allclose(result1.data.lon, result2.data.lon)
    assert_allclose(result1.data.lat, result2.data.lat)
    assert_allclose(result0.data.lon, result1.data.lon)
    assert_allclose(result0.data.lat, result1.data.lat)

    # test for pixel units
    with pytest.raises(ValueError):  # noqa: PT011
        gwcs_2d_shift_scale.array_index_to_world(x * u.Jy, y * u.Jy)


def test_world_to_pixel_quantity(gwcs_2d_shift_scale, gwcs_2d_shift_scale_quantity):
    skycoord = gwcs_2d_shift_scale.pixel_to_world(x, y)
    result1 = gwcs_2d_shift_scale.world_to_pixel(skycoord)
    result2 = gwcs_2d_shift_scale_quantity.world_to_pixel(skycoord)
    assert_allclose(result1, (x, y))
    assert_allclose(result2, (x, y))


def test_world_to_array_index_quantity(
    gwcs_2d_shift_scale, gwcs_2d_shift_scale_quantity
):
    skycoord = gwcs_2d_shift_scale.pixel_to_world(x, y)
    result0 = gwcs_2d_shift_scale.world_to_pixel(skycoord)
    result1 = gwcs_2d_shift_scale.world_to_array_index(skycoord)
    result2 = gwcs_2d_shift_scale_quantity.world_to_array_index(skycoord)
    assert_allclose(result0, (x, y))
    assert_allclose(result1, (y, x))
    assert_allclose(result2, (y, x))


@pytest.fixture(params=[0, 1])
def sky_ra_dec(request, gwcs_2d_spatial_shift):
    ref_frame = gwcs_2d_spatial_shift.output_frame.reference_frame
    ra, dec = 2, 4
    if request.param == 0:
        sky = coord.SkyCoord(ra * u.deg, dec * u.deg, frame=ref_frame)
    else:
        ra = np.ones((3, 4)) * ra
        dec = np.ones((3, 4)) * dec
        sky = coord.SkyCoord(ra * u.deg, dec * u.deg, frame=ref_frame)
    return sky, ra, dec


def test_world_to_pixel(gwcs_2d_spatial_shift, sky_ra_dec):
    wcsobj = gwcs_2d_spatial_shift
    sky, ra, dec = sky_ra_dec
    assert_allclose(wcsobj.world_to_pixel(sky), wcsobj.invert(ra, dec))


def test_world_to_array_index(gwcs_simple_imaging, sky_ra_dec):
    wcsobj = gwcs_simple_imaging
    sky, ra, dec = sky_ra_dec

    assert_allclose(
        wcsobj.world_to_array_index(sky),
        wcsobj.invert(ra, dec)[::-1],
    )


def test_world_to_pixel_values(gwcs_2d_spatial_shift, sky_ra_dec):
    wcsobj = gwcs_2d_spatial_shift
    sky, ra, dec = sky_ra_dec

    assert_allclose(wcsobj.world_to_pixel_values(ra, dec), wcsobj.invert(ra, dec))


def test_world_to_array_index_values(gwcs_simple_imaging, sky_ra_dec):
    wcsobj = gwcs_simple_imaging
    sky, ra, dec = sky_ra_dec

    assert_allclose(
        wcsobj.world_to_array_index_values(ra, dec),
        wcsobj.invert(ra, dec)[::-1],
    )


def test_ndim_str_frames(gwcs_with_frames_strings):
    wcsobj = gwcs_with_frames_strings
    assert wcsobj.pixel_n_dim == 4
    assert wcsobj.world_n_dim == 3


def test_composite_many_base_frame():
    q_frame_1 = cf.CoordinateFrame(
        name="distance", axes_order=(0,), naxes=1, axes_type="SPATIAL", unit=(u.m,)
    )
    q_frame_2 = cf.CoordinateFrame(
        name="distance", axes_order=(1,), naxes=1, axes_type="SPATIAL", unit=(u.m,)
    )
    frame = cf.CompositeFrame([q_frame_1, q_frame_2])

    wao_classes = frame.world_axis_object_classes

    assert len(wao_classes) == 2
    assert not set(wao_classes.keys()).difference({"SPATIAL", "SPATIAL1"})

    wao_components = frame.world_axis_object_components

    assert len(wao_components) == 2
    assert not {c[0] for c in wao_components}.difference({"SPATIAL", "SPATIAL1"})


def test_coordinate_frame_api():
    forward = m.Linear1D(slope=0.1 * u.deg / u.pix, intercept=0 * u.deg)

    output_frame = cf.CoordinateFrame(
        1, "SPATIAL", (0,), unit=(u.deg,), name="sepframe"
    )
    input_frame = cf.CoordinateFrame(1, "PIXEL", (0,), unit=(u.pix,))

    wcs = gwcs.WCS(
        forward_transform=forward, input_frame=input_frame, output_frame=output_frame
    )

    world = wcs.pixel_to_world(0)
    assert isinstance(world, u.Quantity)

    pixel = wcs.world_to_pixel(world)
    assert isinstance(pixel, float)

    with pytest.raises(TypeError):
        _ = wcs.invert(world)


def test_world_axis_object_components_units(gwcs_3d_identity_units):
    from astropy.wcs.wcsapi.high_level_api import high_level_objects_to_values

    wcs = gwcs_3d_identity_units
    world = wcs.pixel_to_world(1, 1, 1)

    values = high_level_objects_to_values(*world, low_level_wcs=wcs)

    expected_values = [
        world[0].spherical.lon.to_value(wcs.output_frame.unit[0]),
        world[0].spherical.lon.to_value(wcs.output_frame.unit[1]),
        world[1].to_value(wcs.output_frame.unit[2]),
    ]

    assert not any(isinstance(o, u.Quantity) for o in values)
    np.testing.assert_allclose(values, expected_values)


def test_mismatched_high_level_types(gwcs_3d_identity_units):
    wcs = gwcs_3d_identity_units

    with pytest.raises(
        TypeError,
        match=(
            "Invalid types were passed.*(tuple, SpectralCoord)"
            ".*(SkyCoord, SpectralCoord).*"
        ),
    ):
        wcs.invert((1 * u.deg, 2 * u.deg), coord.SpectralCoord(10 * u.nm))

    # Oh astropy why do you make us do this
    with pytest.raises(
        TypeError,
        match="Invalid types were passed.*got.*Quantity.*expected.*SpectralCoord.*",
    ):
        wcs.invert(coord.SkyCoord(1 * u.deg, 2 * u.deg), 10 * u.nm)


def test_no_input_frame(gwcs_simple_2d):
    """Test running the API on the WCS with no input frame."""
    assert (np.array([2]), np.array([-1])) == gwcs_simple_2d.world_to_pixel_values(
        np.array([3]), np.array([1])
    )
    assert (np.array([4]), np.array([3])) == gwcs_simple_2d.pixel_to_world_values(
        np.array([3]), np.array([1])
    )


def test_empty_output_frame(gwcs_empty_output_2d):
    """Test running the API on the WCS with an empty output frame."""
    assert (np.array([3]), np.array([1])) == gwcs_empty_output_2d.pixel_to_world_values(
        np.array([2]), np.array([-1])
    )
    assert (
        np.array([2]),
        np.array([-1]),
    ) == gwcs_empty_output_2d.world_to_pixel_values(np.array([3]), np.array([1]))



"""
ASDF tags for spectroscopy related models.

"""

from asdf_astropy.converters.transform.core import (
    TransformConverterBase,
    parameter_to_value,
)
from astropy import units as u

__all__ = [
    "GratingEquationConverter",
    "SellmeierGlassConverter",
    "SellmeierZemaxConverter",
    "Snell3DConverter",
]


class SellmeierGlassConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/sellmeier_glass-*",)
    types = ("gwcs.spectroscopy.SellmeierGlass",)

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.spectroscopy import SellmeierGlass

        return SellmeierGlass(node["B_coef"], node["C_coef"])

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {
            "B_coef": parameter_to_value(model.B_coef),
            "C_coef": parameter_to_value(model.C_coef),
        }


class SellmeierZemaxConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/sellmeier_zemax-*",)
    types = ("gwcs.spectroscopy.SellmeierZemax",)

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.spectroscopy import SellmeierZemax

        return SellmeierZemax(
            node["temperature"],
            node["ref_temperature"],
            node["ref_pressure"],
            node["pressure"],
            node["B_coef"],
            node["C_coef"],
            node["D_coef"],
            node["E_coef"],
        )

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {
            "B_coef": parameter_to_value(model.B_coef),
            "C_coef": parameter_to_value(model.C_coef),
            "D_coef": parameter_to_value(model.D_coef),
            "E_coef": parameter_to_value(model.E_coef),
            "temperature": parameter_to_value(model.temperature),
            "ref_temperature": parameter_to_value(model.ref_temperature),
            "pressure": parameter_to_value(model.pressure),
            "ref_pressure": parameter_to_value(model.ref_pressure),
        }


class Snell3DConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/snell3d-*",)
    types = ("gwcs.spectroscopy.Snell3D",)

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.spectroscopy import Snell3D

        return Snell3D()

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {}


class GratingEquationConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/grating_equation-*",)
    types = (
        "gwcs.spectroscopy.AnglesFromGratingEquation3D",
        "gwcs.spectroscopy.WavelengthFromGratingEquation",
    )

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.spectroscopy import (
            AnglesFromGratingEquation3D,
            WavelengthFromGratingEquation,
        )

        groove_density = node["groove_density"]
        order = node["order"]
        output = node["output"]
        if output == "wavelength":
            model = WavelengthFromGratingEquation(
                groove_density=groove_density, spectral_order=order
            )
        elif output == "angle":
            model = AnglesFromGratingEquation3D(
                groove_density=groove_density, spectral_order=order
            )
        else:
            msg = f"Can't create a GratingEquation model with output {output}"
            raise ValueError(msg)
        return model

    def to_yaml_tree_transform(self, model, tag, ctx):
        from gwcs.spectroscopy import (
            AnglesFromGratingEquation3D,
            WavelengthFromGratingEquation,
        )

        if model.groove_density.unit is not None:
            groove_density = u.Quantity(
                model.groove_density.value, unit=model.groove_density.unit
            )
        else:
            groove_density = model.groove_density.value
        node = {"order": model.spectral_order.value, "groove_density": groove_density}
        if isinstance(model, AnglesFromGratingEquation3D):
            node["output"] = "angle"
        elif isinstance(model, WavelengthFromGratingEquation):
            node["output"] = "wavelength"
        else:
            msg = f"Can't serialize an instance of {model.__class__.__name__}"
            raise TypeError(msg)
        return node



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import importlib.resources

from asdf.extension import Extension, ManifestExtension

from .converters.geometry import DirectionCosinesConverter, SphericalCartesianConverter
from .converters.selector import LabelMapperConverter, RegionsSelectorConverter
from .converters.spectroscopy import (
    GratingEquationConverter,
    SellmeierGlassConverter,
    SellmeierZemaxConverter,
    Snell3DConverter,
)
from .converters.wcs import (
    CelestialFrameConverter,
    CompositeFrameConverter,
    FITSImagingWCSConverter,
    Frame2DConverter,
    FrameConverter,
    SpectralFrameConverter,
    StepConverter,
    StokesFrameConverter,
    TemporalFrameConverter,
    WCSConverter,
)

WCS_MODEL_CONVERTERS = [
    CelestialFrameConverter(),
    CompositeFrameConverter(),
    FrameConverter(),
    Frame2DConverter(),
    SpectralFrameConverter(),
    StepConverter(),
    StokesFrameConverter(),
    TemporalFrameConverter(),
    WCSConverter(),
    LabelMapperConverter(),
    RegionsSelectorConverter(),
    GratingEquationConverter(),
    SellmeierGlassConverter(),
    SellmeierZemaxConverter(),
    Snell3DConverter(),
    DirectionCosinesConverter(),
    SphericalCartesianConverter(),
    FITSImagingWCSConverter(),
]

# The order here is important; asdf will prefer to use extensions
# that occur earlier in the list.
WCS_MANIFEST_URIS = [
    # `importlib.resources.files.iterdir()` regturns a generator of Path objects
    f"asdf://asdf-format.org/astronomy/gwcs/manifests/{path.stem}"  # type: ignore[attr-defined]
    for path in sorted(  # type: ignore[type-var]
        (
            importlib.resources.files("asdf_wcs_schemas.resources") / "manifests"
        ).iterdir(),
        reverse=True,
    )
]

# 1.0.0 contains multiple versions of the same tag, a bug fixed in
# 1.0.1 so only register 1.0.0 if it's the only available manifest
TRANSFORM_EXTENSIONS = [
    ManifestExtension.from_uri(
        uri,
        legacy_class_names=["gwcs.extension.GWCSExtension"],
        converters=WCS_MODEL_CONVERTERS,
    )
    for uri in WCS_MANIFEST_URIS
    if len(WCS_MANIFEST_URIS) == 1 or "1.0.0" not in uri
]

# if we don't register something for the 1.0.0 extension/manifest
# opening old files will issue AsdfWarning messages stating that
# the file was produced with an extension that is not installed
# As the 1.0.1 and 1.1.0 extensions support all the required tags
# it's not a helpful warning so here we register an 'empty'
# extension for 1.0.0 which doesn't support any tags or types
# but will be installed into asdf preventing the warning
if len(TRANSFORM_EXTENSIONS) > 1:

    class _EmptyExtension(Extension):
        extension_uri = "asdf://asdf-format.org/astronomy/gwcs/extensions/gwcs-1.0.0"
        legacy_class_names = ("gwcs.extension.GWCSExtension",)

    TRANSFORM_EXTENSIONS.append(_EmptyExtension())


def get_extensions():
    """
    Get the gwcs.converters extension.
    This method is registered with the asdf.extensions entry point.
    Returns
    -------
    list of asdf.extension.Extension
    """
    return TRANSFORM_EXTENSIONS



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import contextlib
import warnings

import astropy
import astropy.units as u
import numpy as np
import pytest
from astropy import coordinates as coord
from astropy.coordinates import SpectralCoord, StokesCoord
from astropy.modeling import models as m
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time, TimeDelta
from astropy.wcs.wcsapi.fitswcs import CTYPE_TO_UCD1
from numpy.testing import assert_allclose

from gwcs import WCS
from gwcs import coordinate_frames as cf
from gwcs.coordinate_frames._utils import (
    _ALLOWED_UCD_DUPLICATES,
    _ucd1_to_ctype_name_mapping,
)

astropy_version = astropy.__version__

coord_frames = coord.builtin_frames.__all__[:]

# Need to write a better test, using a dict {coord_frame: input_parameters}
# For now remove OffsetFrame, issue #55
with contextlib.suppress(ValueError):
    coord_frames.remove("SkyOffsetFrame")


icrs = cf.CelestialFrame(reference_frame=coord.ICRS(), axes_order=(0, 1))
detector = cf.Frame2D(name="detector", axes_order=(0, 1))
focal = cf.Frame2D(name="focal", axes_order=(0, 1), unit=(u.m, u.m))

spec1 = cf.SpectralFrame(
    name="freq",
    unit=[
        u.Hz,
    ],
    axes_order=(2,),
)
spec2 = cf.SpectralFrame(
    name="wave",
    unit=[
        u.m,
    ],
    axes_order=(2,),
    axes_names=("lambda",),
)
spec3 = cf.SpectralFrame(
    name="energy",
    unit=[
        u.J,
    ],
    axes_order=(2,),
)
spec4 = cf.SpectralFrame(
    name="pixel",
    unit=[
        u.pix,
    ],
    axes_order=(2,),
)
spec5 = cf.SpectralFrame(
    name="speed",
    unit=[
        u.m / u.s,
    ],
    axes_order=(2,),
)

comp1 = cf.CompositeFrame([icrs, spec1])
comp2 = cf.CompositeFrame([focal, spec2])
comp3 = cf.CompositeFrame([icrs, spec3])
comp4 = cf.CompositeFrame([icrs, spec4])
comp5 = cf.CompositeFrame([icrs, spec5])
comp = cf.CompositeFrame([comp1, cf.SpectralFrame(axes_order=(3,), unit=(u.m,))])

xscalar = 1
yscalar = 2
xarr = np.arange(5)
yarr = np.arange(5)

inputs2 = [(xscalar, yscalar), (xarr, yarr)]
inputs1 = [xscalar, xarr]
inputs3 = [(xscalar, yscalar, xscalar), (xarr, yarr, xarr)]


def test_units():
    assert comp1.unit == (u.deg, u.deg, u.Hz)
    assert comp2.unit == (u.m, u.m, u.m)
    assert comp3.unit == (u.deg, u.deg, u.J)
    assert comp4.unit == (u.deg, u.deg, u.pix)
    assert comp5.unit == (u.deg, u.deg, u.m / u.s)
    assert comp.unit == (u.deg, u.deg, u.Hz, u.m)


# These two functions fake the old methods on CoordinateFrame to reduce the
# amount of refactoring that needed doing in these tests.
def coordinates(*inputs, frame):
    return frame.to_high_level_coordinates(*inputs)


def coordinate_to_quantity(*inputs, frame):
    results = frame.from_high_level_coordinates(*inputs)
    if not isinstance(results, list):
        results = [results]
    return [r << unit for r, unit in zip(results, frame.unit, strict=False)]


@pytest.mark.parametrize("inputs", inputs2)
def test_coordinates_spatial(inputs):
    sky_coord = coordinates(*inputs, frame=icrs)
    assert isinstance(sky_coord, coord.SkyCoord)
    assert_allclose((sky_coord.ra.value, sky_coord.dec.value), inputs)
    focal_coord = coordinates(*inputs, frame=focal)
    assert_allclose([coord.value for coord in focal_coord], inputs)
    assert [coord.unit for coord in focal_coord] == [u.m, u.m]


@pytest.mark.parametrize("inputs", inputs1)
def test_coordinates_spectral(inputs):
    wave = coordinates(inputs, frame=spec2)
    assert_allclose(wave.value, inputs)
    assert wave.unit == "meter"
    assert isinstance(wave, u.Quantity)


@pytest.mark.parametrize("inputs", inputs3)
def test_coordinates_composite(inputs):
    frame = cf.CompositeFrame([icrs, spec2])
    result = coordinates(*inputs, frame=frame)
    assert isinstance(result[0], coord.SkyCoord)
    assert_allclose((result[0].ra.value, result[0].dec.value), inputs[:2])
    assert_allclose(result[1].value, inputs[2])


def test_coordinates_composite_order():
    time = cf.TemporalFrame(
        Time("2011-01-01T00:00:00"),
        name="time",
        unit=[
            u.s,
        ],
        axes_order=(0,),
    )
    dist = cf.CoordinateFrame(
        name="distance",
        naxes=1,
        axes_type=["SPATIAL"],
        unit=[
            u.m,
        ],
        axes_order=(1,),
    )
    frame = cf.CompositeFrame([time, dist])
    result = coordinates(0, 0, frame=frame)
    assert result[0] == Time("2011-01-01T00:00:00")
    assert u.allclose(result[1], 0 * u.m)


def test_bare_baseframe():
    # This is a regression test for the following call:
    frame = cf.CoordinateFrame(1, "SPATIAL", (0,), unit=(u.km,))
    quantity = coordinate_to_quantity(1 * u.m, frame=frame)
    assert u.allclose(quantity, 1 * u.m)

    # Now also setup the same situation through the whole call stack to be safe.
    w = WCS(
        forward_transform=m.Tabular1D(
            points=np.arange(10) * u.pix, lookup_table=np.arange(10) * u.km
        ),
        output_frame=frame,
        input_frame=cf.CoordinateFrame(
            1, "PIXEL", (0,), unit=(u.pix,), name="detector_frame"
        ),
    )
    assert u.allclose(w.world_to_pixel(0 * u.km), 0)


@pytest.mark.parametrize(("frame"), coord_frames)
def test_celestial_attributes_length(frame):
    """
    Test getting default values for
    CelestialFrame attributes from reference_frame.
    """
    fr = getattr(coord, frame)
    if issubclass(fr.__class__, coord.BaseCoordinateFrame):
        cel = cf.CelestialFrame(reference_frame=fr())
        assert (
            len(cel.axes_names)
            == len(cel.axes_type)
            == len(cel.unit)
            == len(cel.axes_order)
            == cel.naxes
        )


def test_axes_type():
    assert icrs.axes_type == ("SPATIAL", "SPATIAL")
    assert spec1.axes_type == ("SPECTRAL",)
    assert detector.axes_type == ("SPATIAL", "SPATIAL")
    assert focal.axes_type == ("SPATIAL", "SPATIAL")


def test_length_attributes():
    with pytest.raises(ValueError):  # noqa: PT011
        cf.CoordinateFrame(
            naxes=2, unit=(u.deg), axes_type=("SPATIAL", "SPATIAL"), axes_order=(0, 1)
        )

    with pytest.raises(ValueError):  # noqa: PT011
        cf.CoordinateFrame(
            naxes=2, unit=(u.deg, u.deg), axes_type=("SPATIAL",), axes_order=(0, 1)
        )

    with pytest.raises(ValueError):  # noqa: PT011
        cf.CoordinateFrame(
            naxes=2,
            unit=(u.deg, u.deg),
            axes_type=("SPATIAL", "SPATIAL"),
            axes_order=(0,),
        )


def test_base_coordinate():
    frame = cf.CoordinateFrame(
        naxes=2, axes_type=("SPATIAL", "SPATIAL"), axes_order=(0, 1)
    )
    assert frame.name == "CoordinateFrame"
    frame = cf.CoordinateFrame(
        name="CustomFrame",
        naxes=2,
        axes_type=("SPATIAL", "SPATIAL"),
        axes_order=(0, 1),
        unit=(u.deg, u.arcsec),
    )
    assert frame.name == "CustomFrame"
    frame.name = "DeLorean"
    assert frame.name == "DeLorean"

    q1, q2 = coordinate_to_quantity(12 * u.deg, 3 * u.arcsec, frame=frame)
    assert_quantity_allclose(q1, 12 * u.deg)
    assert_quantity_allclose(q2, 3 * u.arcsec)

    q1, q2 = coordinate_to_quantity(*(12 * u.deg, 3 * u.arcsec), frame=frame)
    assert_quantity_allclose(q1, 12 * u.deg)
    assert_quantity_allclose(q2, 3 * u.arcsec)


def test_temporal_relative():
    t = cf.TemporalFrame(reference_frame=Time("2018-01-01T00:00:00"), unit=u.s)
    assert coordinates(10, frame=t) == Time("2018-01-01T00:00:00") + 10 * u.s
    assert coordinates(10 * u.s, frame=t) == Time("2018-01-01T00:00:00") + 10 * u.s

    a = coordinates(np.array((10, 20)), frame=t)
    assert a[0] == Time("2018-01-01T00:00:00") + 10 * u.s
    assert a[1] == Time("2018-01-01T00:00:00") + 20 * u.s

    t = cf.TemporalFrame(reference_frame=Time("2018-01-01T00:00:00"))
    assert coordinates(10 * u.s, frame=t) == Time("2018-01-01T00:00:00") + 10 * u.s
    assert (
        coordinates(TimeDelta(10, format="sec"), frame=t)
        == Time("2018-01-01T00:00:00") + 10 * u.s
    )

    a = coordinates(np.array((10, 20)) * u.s, frame=t)
    assert a[0] == Time("2018-01-01T00:00:00") + 10 * u.s
    assert a[1] == Time("2018-01-01T00:00:00") + 20 * u.s


@pytest.mark.parametrize(
    "inp",
    [
        (coord.SkyCoord(10 * u.deg, 20 * u.deg, frame=coord.ICRS),),
        # This is the same as 10,20 in ICRS
        (coord.SkyCoord(119.26936774, -42.79039286, unit=u.deg, frame="galactic"),),
    ],
)
def test_coordinate_to_quantity_celestial(inp):
    cel = cf.CelestialFrame(reference_frame=coord.ICRS(), axes_order=(0, 1))

    lon, lat = coordinate_to_quantity(*inp, frame=cel)
    assert_quantity_allclose(lon, 10 * u.deg)
    assert_quantity_allclose(lat, 20 * u.deg)

    with pytest.raises(ValueError):  # noqa: PT011
        coordinate_to_quantity(10 * u.deg, 2 * u.deg, 3 * u.deg, frame=cel)

    with pytest.raises(ValueError):  # noqa: PT011
        coordinate_to_quantity((1, 2), frame=cel)


@pytest.mark.parametrize(
    "inp",
    [
        (SpectralCoord(100 * u.nm),),
        (SpectralCoord(0.1 * u.um),),
    ],
)
def test_coordinate_to_quantity_spectral(inp):
    spec = cf.SpectralFrame(unit=u.nm, axes_order=(1,))
    wav = coordinate_to_quantity(*inp, frame=spec)
    assert_quantity_allclose(wav, 100 * u.nm)


@pytest.mark.parametrize(
    "inp",
    [
        (Time("2011-01-01T00:00:10"),),
    ],
)
def test_coordinate_to_quantity_temporal(inp):
    temp = cf.TemporalFrame(reference_frame=Time("2011-01-01T00:00:00"), unit=u.s)

    t = coordinate_to_quantity(*inp, frame=temp)

    assert_quantity_allclose(t, 10 * u.s)


@pytest.mark.parametrize(
    "inp",
    [
        (
            SpectralCoord(211 * u.AA),
            Time("2011-01-01T00:00:00"),
            coord.SkyCoord(0, 0, unit=u.arcsec),
        ),
    ],
)
def test_coordinate_to_quantity_composite(inp):
    # Composite
    wave_frame = cf.SpectralFrame(axes_order=(0,), unit=u.AA)
    time_frame = cf.TemporalFrame(
        axes_order=(1,), unit=u.s, reference_frame=Time("2011-01-01T00:00:00")
    )
    sky_frame = cf.CelestialFrame(axes_order=(2, 3), reference_frame=coord.ICRS())

    comp = cf.CompositeFrame([wave_frame, time_frame, sky_frame])

    coords = coordinate_to_quantity(*inp, frame=comp)

    expected = (211 * u.AA, 0 * u.s, 0 * u.arcsec, 0 * u.arcsec)
    for output, exp in zip(coords, expected, strict=False):
        assert_quantity_allclose(output, exp)


def test_coordinate_to_quantity_composite_split():
    inp = (
        SpectralCoord(211 * u.AA),
        coord.SkyCoord(0, 0, unit=u.arcsec),
        Time("2011-01-01T00:00:00"),
    )

    # Composite
    wave_frame = cf.SpectralFrame(axes_order=(1,), unit=u.AA)
    sky_frame = cf.CelestialFrame(axes_order=(2, 0), reference_frame=coord.ICRS())
    time_frame = cf.TemporalFrame(
        axes_order=(3,), unit=u.s, reference_frame=Time("2011-01-01T00:00:00")
    )

    comp = cf.CompositeFrame([wave_frame, sky_frame, time_frame])

    coords = coordinate_to_quantity(*inp, frame=comp)

    expected = (0 * u.arcsec, 211 * u.AA, 0 * u.arcsec, 0 * u.s)
    for output, exp in zip(coords, expected, strict=False):
        assert_quantity_allclose(output, exp)


def test_stokes_frame():
    sf = cf.StokesFrame()

    assert coordinates(1, frame=sf) == "I"
    assert coordinates(1 * u.one, frame=sf) == "I"
    assert coordinate_to_quantity(StokesCoord("I"), frame=sf) == 1 * u.one
    assert coordinate_to_quantity(StokesCoord(1), frame=sf) == 1 * u.one


def test_coordinate_to_quantity_frame2d_composite():
    inp = (SpectralCoord(211 * u.AA), Time("2011-01-01T00:00:00"), 0 * u.one, 0 * u.one)
    wave_frame = cf.SpectralFrame(axes_order=(0,), unit=u.AA)
    time_frame = cf.TemporalFrame(
        axes_order=(1,), unit=u.s, reference_frame=Time("2011-01-01T00:00:00")
    )

    frame2d = cf.Frame2D(name="intermediate", axes_order=(2, 3), unit=(u.one, u.one))

    comp = cf.CompositeFrame([wave_frame, time_frame, frame2d])

    coords = coordinate_to_quantity(*inp, frame=comp)

    expected = (211 * u.AA, 0 * u.s, 0 * u.one, 0 * u.one)
    for output, exp in zip(coords, expected, strict=False):
        assert_quantity_allclose(output, exp)


def test_coordinate_to_quantity_frame_2d():
    frame = cf.Frame2D(unit=(u.one, u.arcsec))
    inp = (1 * u.one, 2 * u.arcsec)
    expected = (1 * u.one, 2 * u.arcsec)
    result = coordinate_to_quantity(*inp, frame=frame)
    for output, exp in zip(result, expected, strict=False):
        assert_quantity_allclose(output, exp)


def test_coordinate_to_quantity_error():
    frame = cf.Frame2D(unit=(u.one, u.arcsec))
    with pytest.raises(ValueError):  # noqa: PT011
        coordinate_to_quantity(1, frame=frame)

    with pytest.raises(ValueError):  # noqa: PT011
        coordinate_to_quantity((1, 1), 2, frame=frame)

    frame = cf.TemporalFrame(reference_frame=Time([], format="isot"), unit=u.s)
    with pytest.raises(ValueError):  # noqa: PT011
        coordinate_to_quantity(1, frame=frame)


def test_axis_physical_types():
    assert icrs.axis_physical_types == ("pos.eq.ra", "pos.eq.dec")
    assert spec1.axis_physical_types == ("em.freq",)
    assert spec2.axis_physical_types == ("em.wl",)
    assert spec3.axis_physical_types == ("em.energy",)
    assert spec4.axis_physical_types == ("custom:unknown",)
    assert spec5.axis_physical_types == ("spect.dopplerVeloc",)
    assert comp1.axis_physical_types == ("pos.eq.ra", "pos.eq.dec", "em.freq")
    assert comp2.axis_physical_types == ("custom:x", "custom:y", "em.wl")
    assert comp3.axis_physical_types == ("pos.eq.ra", "pos.eq.dec", "em.energy")
    assert comp.axis_physical_types == ("pos.eq.ra", "pos.eq.dec", "em.freq", "em.wl")

    spec6 = cf.SpectralFrame(
        name="waven",
        axes_order=(1,),
        axis_physical_types="em.wavenumber",
        unit=u.Unit(1),
    )
    assert spec6.axis_physical_types == ("em.wavenumber",)

    t = cf.TemporalFrame(reference_frame=Time("2018-01-01T00:00:00"), unit=u.s)
    assert t.axis_physical_types == ("time",)

    fr2d = cf.Frame2D(name="d", axes_names=("x", "y"))
    assert fr2d.axis_physical_types == ("custom:x", "custom:y")

    fr2d = cf.Frame2D(name="d", axes_names=None)
    assert fr2d.axis_physical_types == ("custom:SPATIAL", "custom:SPATIAL")

    fr2d = cf.Frame2D(name="d", axis_physical_types=("pos.x", "pos.y"))
    assert fr2d.axis_physical_types == ("custom:pos.x", "custom:pos.y")

    with pytest.raises(ValueError):  # noqa: PT011
        cf.CelestialFrame(
            reference_frame=coord.ICRS(), axis_physical_types=("pos.eq.ra",)
        )

    fr = cf.CelestialFrame(
        reference_frame=coord.ICRS(), axis_physical_types=("ra", "dec")
    )
    assert fr.axis_physical_types == ("custom:ra", "custom:dec")

    fr = cf.CelestialFrame(reference_frame=coord.BarycentricTrueEcliptic())
    assert fr.axis_physical_types == ("pos.ecliptic.lon", "pos.ecliptic.lat")

    frame = cf.CoordinateFrame(
        name="custom_frame",
        axes_type=("SPATIAL",),
        axes_order=(0,),
        axis_physical_types="length",
        axes_names="x",
        naxes=1,
    )
    assert frame.axis_physical_types == ("custom:length",)
    frame = cf.CoordinateFrame(
        name="custom_frame",
        axes_type=("SPATIAL",),
        axes_order=(0,),
        axis_physical_types=("length",),
        axes_names="x",
        naxes=1,
    )
    assert frame.axis_physical_types == ("custom:length",)
    with pytest.raises(ValueError):  # noqa: PT011
        cf.CoordinateFrame(
            name="custom_frame",
            axes_type=("SPATIAL",),
            axes_order=(0,),
            axis_physical_types=("length", "length"),
            naxes=1,
        )


def test_base_frame():
    with pytest.raises(ValueError):  # noqa: PT011
        cf.CoordinateFrame(
            name="custom_frame",
            axes_type=("SPATIAL",),
            naxes=1,
            axes_order=(0,),
            axes_names=("x", "y"),
        )
    frame = cf.CoordinateFrame(
        name="custom_frame",
        axes_type=("SPATIAL",),
        axes_order=(0,),
        axes_names="x",
        naxes=1,
    )
    assert frame.naxes == 1
    assert frame.axes_names == ("x",)

    coordinate_to_quantity(1 * u.one, frame=frame)


def test_ucd1_to_ctype_not_out_of_sync(caplog):
    """
    Test that our code is not out-of-sync with ``astropy``'s definition of
    ``CTYPE_TO_UCD1`` and our dictionary of allowed duplicates.

    If this test is failing, update ``coordinate_frames._ALLOWED_UCD_DUPLICATES``
    dictionary with new types defined in ``astropy``'s ``CTYPE_TO_UCD1``.

    """
    _ucd1_to_ctype_name_mapping(
        ctype_to_ucd=CTYPE_TO_UCD1, allowed_ucd_duplicates=_ALLOWED_UCD_DUPLICATES
    )

    assert len(caplog.record_tuples) == 0


def test_ucd1_to_ctype():
    new_ctype_to_ucd = {
        "RPT1": "new.repeated.type",
        "RPT2": "new.repeated.type",
        "RPT3": "new.repeated.type",
    }

    ctype_to_ucd = dict(**CTYPE_TO_UCD1, **new_ctype_to_ucd)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        inv_map = _ucd1_to_ctype_name_mapping(
            ctype_to_ucd=ctype_to_ucd, allowed_ucd_duplicates=_ALLOWED_UCD_DUPLICATES
        )
        assert len(w) == 1
        assert issubclass(w[0].category, UserWarning)
        assert "Found unsupported duplicate physical type" in str(w[0].message)

    for k, v in _ALLOWED_UCD_DUPLICATES.items():
        assert inv_map.get(k, "") == v

    for k, v in inv_map.items():
        assert ctype_to_ucd[v] == k

    assert inv_map["new.repeated.type"] in new_ctype_to_ucd


def test_celestial_ordering():
    c1 = cf.CelestialFrame(
        reference_frame=coord.ICRS(),
        axes_order=(0, 1),
        axes_names=("lon", "lat"),
        unit=(u.deg, u.arcsec),
        axis_physical_types=("custom:lon", "custom:lat"),
    )
    c2 = cf.CelestialFrame(
        reference_frame=coord.ICRS(),
        axes_order=(1, 0),
        axes_names=("lon", "lat"),
        unit=(u.deg, u.arcsec),
        axis_physical_types=("custom:lon", "custom:lat"),
    )

    assert c1.axes_names == ("lon", "lat")
    assert c2.axes_names == ("lat", "lon")

    assert c1.unit == (u.deg, u.arcsec)
    assert c2.unit == (u.arcsec, u.deg)

    assert c1.axis_physical_types == ("custom:lon", "custom:lat")
    assert c2.axis_physical_types == ("custom:lat", "custom:lon")


def test_composite_ordering():
    c1 = cf.CelestialFrame(
        reference_frame=coord.ICRS(),
        axes_order=(1, 0),
        axes_names=("lon", "lat"),
        unit=(u.deg, u.arcsec),
        axis_physical_types=("custom:lon", "custom:lat"),
    )
    spec = cf.SpectralFrame(
        axes_order=(2,),
        axes_names=("spectral",),
        unit=u.AA,
    )
    comp = cf.CompositeFrame([c1, spec])
    assert comp.axes_names == ("lat", "lon", "spectral")
    assert comp.axis_physical_types == ("custom:lat", "custom:lon", "em.wl")
    assert comp.unit == (u.arcsec, u.deg, u.AA)
    assert comp.axes_order == (1, 0, 2)



import io
import warnings

import asdf
import asdf_wcs_schemas
import pytest

import gwcs.extension


@pytest.mark.skipif(
    asdf_wcs_schemas.__version__ < "0.2.0",
    reason="version 0.2 provides the new manifests",
)
def test_empty_extension():
    """
    Test that an empty extension was installed for gwcs 1.0.0
    and that extensions are installed for gwcs 1.0.1 and 1.1.0
    """
    extensions = gwcs.extension.get_extensions()
    assert len(extensions) > 1

    extensions_by_uri = {ext.extension_uri: ext for ext in extensions}

    # check for duplicate uris
    assert len(extensions_by_uri) == len(extensions)

    # check that all 3 versions are installed
    for version in ("1.0.0", "1.0.1", "1.1.0"):
        assert (
            f"asdf://asdf-format.org/astronomy/gwcs/extensions/gwcs-{version}"
            in extensions_by_uri
        )

    # the 1.0.0 extension should support no tags or types
    legacy = extensions_by_uri[
        "asdf://asdf-format.org/astronomy/gwcs/extensions/gwcs-1.0.0"
    ]
    assert len(legacy.tags) == 0
    assert len(legacy.converters) == 0


def test_open_legacy_without_warning():
    """
    Opening a file produced with extension 1.0.0 should not produce any
    warnings because of the empty extension registered for 1.0.0
    """
    asdf_bytes = b"""#ASDF 1.0.0
#ASDF_STANDARD 1.5.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.1.0
asdf_library: !core/software-1.0.0 {author: The ASDF Developers, homepage: 'http://github.com/asdf-format/asdf',
  name: asdf, version: 2.9.2}
history:
  extensions:
  - !core/extension_metadata-1.0.0
    extension_class: asdf.extension._manifest.ManifestExtension
    extension_uri: asdf://asdf-format.org/astronomy/gwcs/extensions/gwcs-1.0.0
    software: !core/software-1.0.0 {name: gwcs, version: 0.18.0}
foo: 1
..."""
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        with asdf.open(io.BytesIO(asdf_bytes)) as af:
            assert af["foo"] == 1



import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

x = [-1, 2, 4, 13]
y = [np.nan, np.nan, 4, np.nan]
y1 = [np.nan, np.nan, 4, np.nan]


@pytest.mark.parametrize(
    (("input_", "output")),
    [((2, 4), (2, 4)), ((100, 200), (np.nan, np.nan)), ((x, x), (y, y))],
)
def test_2d_spatial(gwcs_2d_spatial_shift, input_, output):
    w = gwcs_2d_spatial_shift
    w.bounding_box = ((-0.5, 21), (4, 12))

    assert_array_equal(w.invert(*w(*input_)), output)
    assert_array_equal(
        w.world_to_pixel_values(*w.pixel_to_world_values(*input_)), output
    )
    assert_array_equal(w.world_to_pixel(w.pixel_to_world(*input_)), output)


@pytest.mark.parametrize(
    (("input_", "output")),
    [((2, 4), (2, 4)), ((100, 200), (np.nan, np.nan)), ((x, x), (y, y))],
)
def test_2d_spatial_coordinate(gwcs_2d_quantity_shift, input_, output):
    w = gwcs_2d_quantity_shift
    w.bounding_box = ((-0.5, 21), (4, 12))

    assert_array_equal(w.invert(*w(*input_)), output)
    assert_array_equal(
        w.world_to_pixel_values(*w.pixel_to_world_values(*input_)), output
    )
    assert_array_equal(w.world_to_pixel(*w.pixel_to_world(*input_)), output)


@pytest.mark.parametrize(
    (("input_", "output")),
    [((2, 4), (2, 4)), ((100, 200), (np.nan, np.nan)), ((x, x), (y, y))],
)
def test_2d_spatial_coordinate_reordered(gwcs_2d_spatial_reordered, input_, output):
    w = gwcs_2d_spatial_reordered
    w.bounding_box = ((-0.5, 21), (4, 12))

    assert_array_equal(w.invert(*w(*input_)), output)
    assert_array_equal(
        w.world_to_pixel_values(*w.pixel_to_world_values(*input_)), output
    )
    assert_array_equal(w.world_to_pixel(w.pixel_to_world(*input_)), output)


@pytest.mark.parametrize(
    (("input_", "output")), [(2, 2), ((10, 200), (10, np.nan)), (x, (np.nan, 2, 4, 13))]
)
def test_1d_freq(gwcs_1d_freq, input_, output):
    w = gwcs_1d_freq
    w.bounding_box = (-0.5, 21)
    assert_array_equal(w.invert(w(input_)), output)
    assert_array_equal(w.world_to_pixel_values(w.pixel_to_world_values(input_)), output)
    assert_array_equal(w.world_to_pixel(w.pixel_to_world(input_)), output)


@pytest.mark.parametrize(
    (("input_", "output")),
    [
        ((2, 4, 5), (2, 4, 5)),
        ((100, 200, 5), (np.nan, np.nan, np.nan)),
        ((x, x, x), (y1, y1, y1)),
    ],
)
def test_3d_spatial_wave(gwcs_3d_spatial_wave, input_, output):
    w = gwcs_3d_spatial_wave
    w.bounding_box = ((-0.5, 21), (4, 12), (3, 21))

    assert_array_equal(w.invert(*w(*input_)), output)
    assert_array_equal(
        w.world_to_pixel_values(*w.pixel_to_world_values(*input_)), output
    )
    assert_array_equal(w.world_to_pixel(*w.pixel_to_world(*input_)), output)


@pytest.mark.parametrize(
    (("input_", "output")),
    [
        ((1, 2, 3, 4), (1.0, 2.0, 3.0, 4.0)),
        ((100, 3, 3, 3), (np.nan, 3, 3, 3)),
        (
            (x, x, x, x),
            [
                [np.nan, 2.0, 4.0, 13.0],
                [np.nan, 2.0, 4.0, 13.0],
                [np.nan, 2.0, 4.0, 13.0],
                [np.nan, 2.0, 4.0, np.nan],
            ],
        ),
    ],
)
def test_gwcs_spec_cel_time_4d(gwcs_spec_cel_time_4d, input_, output):
    w = gwcs_spec_cel_time_4d

    assert_allclose(w.invert(*w(*input_, with_bounding_box=False)), output, atol=1e-8)



import contextlib

import numpy as np
from astropy import time
from astropy import units as u

from ._core import CoordinateFrame

__all__ = ["TemporalFrame"]


class TemporalFrame(CoordinateFrame):
    """
    A coordinate frame for time axes.

    Parameters
    ----------
    reference_frame : `~astropy.time.Time`
        A Time object which holds the time scale and format.
        If data is provided, it is the time zero point.
        To not set a zero point for the frame initialize ``reference_frame``
        with an empty list.
    unit : str or `~astropy.units.Unit`
        Time unit.
    axes_names : str
        Time axis name.
    axes_order : tuple or int
        A dimension in the data that corresponds to this axis.
    name : str
        Name for this frame.
    """

    def __init__(
        self,
        reference_frame,
        unit=u.s,
        axes_order=(0,),
        axes_names=None,
        name=None,
        axis_physical_types=None,
    ):
        axes_names = (
            axes_names
            or f"{reference_frame.format}({reference_frame.scale}; "
            f"{reference_frame.location}"
        )

        pht = axis_physical_types or self._default_axis_physical_types()

        super().__init__(
            naxes=1,
            axes_type="TIME",
            axes_order=axes_order,
            axes_names=axes_names,
            reference_frame=reference_frame,
            unit=unit,
            name=name,
            axis_physical_types=pht,
        )
        self._attrs = {}
        for a in self.reference_frame.info._represent_as_dict_extra_attrs:
            with contextlib.suppress(AttributeError):
                self._attrs[a] = getattr(self.reference_frame, a)

    def _default_axis_physical_types(self):
        return ("time",)

    def _convert_to_time(self, dt, *, unit, **kwargs):
        if (
            not isinstance(dt, time.TimeDelta) and isinstance(dt, time.Time)
        ) or isinstance(self.reference_frame.value, np.ndarray):
            return time.Time(dt, **kwargs)

        if not hasattr(dt, "unit"):
            dt = dt * unit

        return self.reference_frame + dt

    @property
    def world_axis_object_classes(self):
        comp = (
            time.Time,
            (),
            {"unit": self.unit[0], **self._attrs},
            self._convert_to_time,
        )

        return {"temporal": comp}

    @property
    def _native_world_axis_object_components(self):
        if isinstance(self.reference_frame.value, np.ndarray):
            return [("temporal", 0, "value")]

        def offset_from_time_and_reference(time):
            return (time - self.reference_frame).sec

        return [("temporal", 0, offset_from_time_and_reference)]



import sys
import warnings

import numpy as np
import numpy.linalg as npla
from astropy.modeling.models import (
    Const1D,
    Identity,
    Mapping,
    Polynomial2D,
)
from scipy import linalg

from gwcs.wcstools import grid_from_bounding_box

__all__ = [
    "fit_2D_poly",
    "fix_transform_inputs",
    "make_sampling_grid",
    "reform_poly_coefficients",
    "store_2D_coefficients",
]


def _poly_fit_lu(xin, yin, xout, yout, degree, coord_pow=None):
    # This function fits 2D polynomials to data by writing the normal system
    # of equations and solving it using LU-decomposition. In theory this
    # should be less stable than the SVD method used by numpy's lstsq or
    # astropy's LinearLSQFitter because the condition of the normal matrix
    # is squared compared to the direct matrix. However, in practice,
    # in our (Mihai Cara) tests of fitting WCS distortions, solving the
    # normal system proved to be significantly more accurate, efficient,
    # and stable than SVD.
    #
    # coord_pow - a dictionary used to store powers of coordinate arrays
    #    of the form x**p * y**q used to build the pseudo-Vandermonde matrix.
    #    This improves efficiency especially when fitting multiple degrees
    #    on the same coordinate grid in _fit_2D_poly by reusing computed
    #    powers.
    powers = [
        (i, j) for i in range(degree + 1) for j in range(degree + 1 - i) if i + j > 0
    ]
    if coord_pow is None:
        coord_pow = {}

    nterms = len(powers)

    flt_type = np.longdouble

    # allocate array for the coefficients of the system of equations (a*x=b):
    a = np.empty((nterms, nterms), dtype=flt_type)
    bx = np.empty(nterms, dtype=flt_type)
    by = np.empty(nterms, dtype=flt_type)

    xout = xout.ravel()
    yout = yout.ravel()

    x = np.asarray(xin.ravel(), dtype=flt_type)
    y = np.asarray(yin.ravel(), dtype=flt_type)

    # pseudo_vander - a reduced Vandermonde matrix for 2D polynomials
    # that has only terms x^i * y^j with powers i, j that satisfy:
    # 0 < i + j <= degree.
    pseudo_vander = np.empty((x.size, nterms), dtype=float)

    def pow2(p, q):
        # computes product of powers of coordinate arrays (x**p) * (y**q)
        # in an efficient way avoiding unnecessary array copying
        # and/or raising to power
        if (p, q) in coord_pow:
            return coord_pow[(p, q)]
        if p == 0:
            arr = y**q if q > 1 else y
        elif q == 0:
            arr = x**p if p > 1 else x
        else:
            xp = x if p == 1 else x**p
            yq = y if q == 1 else y**q
            arr = xp * yq
        coord_pow[(p, q)] = arr
        return arr

    for i in range(nterms):
        pi, qi = powers[i]
        coord_pq = pow2(pi, qi)
        pseudo_vander[:, i] = coord_pq
        bx[i] = np.sum(xout * coord_pq, dtype=flt_type)
        by[i] = np.sum(yout * coord_pq, dtype=flt_type)

        for j in range(i, nterms):
            pj, qj = powers[j]
            coord_pq = pow2(pi + pj, qi + qj)
            a[i, j] = np.sum(coord_pq, dtype=flt_type)
            a[j, i] = a[i, j]

    with warnings.catch_warnings(record=True):
        warnings.simplefilter("error", category=linalg.LinAlgWarning)
        try:
            lu_piv = linalg.lu_factor(a)
            poly_coeff_x = linalg.lu_solve(lu_piv, bx).astype(float)
            poly_coeff_y = linalg.lu_solve(lu_piv, by).astype(float)
        except (ValueError, linalg.LinAlgWarning, np.linalg.LinAlgError) as e:
            msg = f"Failed to fit SIP. Reported error:\n{e.args[0]}"
            raise np.linalg.LinAlgError(msg) from e

    if not np.all(np.isfinite([poly_coeff_x, poly_coeff_y])):
        msg = "Failed to fit SIP. Computed coefficients are not finite."
        raise np.linalg.LinAlgError(msg)

    cond = np.linalg.cond(a.astype(float))

    fitx = np.dot(pseudo_vander, poly_coeff_x)
    fity = np.dot(pseudo_vander, poly_coeff_y)

    dist = np.sqrt((xout - fitx) ** 2 + (yout - fity) ** 2)
    max_resid = dist.max()

    return poly_coeff_x, poly_coeff_y, max_resid, powers, cond


def fit_2D_poly(
    degree,
    max_error,
    plate_scale,
    xin,
    yin,
    xout,
    yout,
    xind,
    yind,
    xoutd,
    youtd,
    verbose=False,
):
    """
    Fit a pair of ordinary 2D polynomials to the supplied transform.

    """
    # The case of one pass with the specified polynomial degree
    if degree is None:
        deglist = list(range(1, 10))
    elif hasattr(degree, "__iter__"):
        deglist = sorted(map(int, degree))
        if deglist[0] < 1 or deglist[-1] > 9:
            msg = "Allowed values for SIP degree are [1...9]"
            raise ValueError(msg)
    else:
        degree = int(degree)
        if degree < 1 or degree > 9:
            msg = "Allowed values for SIP degree are [1...9]"
            raise ValueError(msg)
        deglist = [degree]

    single_degree = len(deglist) == 1

    fit_error = np.inf
    if verbose and not single_degree:
        sys.stdout.write(f"Maximum specified SIP approximation error: {max_error}")
    max_error *= plate_scale

    fit_warning_msg = "Failed to achieve requested SIP approximation accuracy."

    # Fit lowest degree SIP first.
    coord_pow = {}  # hold coordinate arrays powers for optimization purpose
    for deg in deglist:
        try:
            cfx_i, cfy_i, fit_error_i, powers_i, cond = _poly_fit_lu(
                xin, yin, xout, yout, degree=deg, coord_pow=coord_pow
            )
            if verbose and not single_degree:
                sys.stdout.write(
                    f"   - SIP degree: {deg}. "
                    f"Maximum residual: {fit_error_i / plate_scale:.5g}"
                )

        except np.linalg.LinAlgError:
            if single_degree:
                # Nothing to do if failure is for the lowest degree
                raise
            # Keep results from the previous iteration. Discard current fit
            break

        if not np.isfinite(cond):
            # Ill-conditioned system
            if single_degree:
                warnings.warn("The fit may be poorly conditioned.", stacklevel=2)
                cfx = cfx_i
                cfy = cfy_i
                fit_error = fit_error_i
                powers = powers_i
            break

        if fit_error_i >= fit_error:
            # Accuracy does not improve. Likely ill-conditioned system
            break

        cfx = cfx_i
        cfy = cfy_i
        powers = powers_i

        fit_error = fit_error_i

        if fit_error <= max_error:
            # Requested accuracy has been achieved
            fit_warning_msg = None
            break

        # Continue to the next degree

    fit_poly_x = Polynomial2D(degree=deg, c0_0=0.0)
    fit_poly_y = Polynomial2D(degree=deg, c0_0=0.0)
    for cx, cy, (p, q) in zip(cfx, cfy, powers, strict=False):
        setattr(fit_poly_x, f"c{p:1d}_{q:1d}", cx)
        setattr(fit_poly_y, f"c{p:1d}_{q:1d}", cy)

    if fit_warning_msg:
        warnings.warn(fit_warning_msg, linalg.LinAlgWarning, stacklevel=2)

    if fit_error <= max_error or single_degree:
        # Check to see if double sampling meets error requirement.
        max_resid = _compute_distance_residual(
            xoutd, youtd, fit_poly_x(xind, yind), fit_poly_y(xind, yind)
        )
        if verbose:
            sys.stdout.write(
                "* Maximum residual, double sampled grid: "
                f"{max_resid / plate_scale:.5g}"
            )

        if max_resid > min(5.0 * fit_error, max_error):
            warnings.warn(
                "Double sampling check FAILED: Sampling may be too coarse for "
                "the distortion model being fitted.",
                stacklevel=2,
            )

        # Residuals on the double-dense grid may be better estimates
        # of the accuracy of the fit. So we report the largest of
        # the residuals (on single- and double-sampled grid) as the fit error:
        fit_error = max(max_resid, fit_error)

    if verbose:
        if single_degree:
            sys.stdout.write(f"Maximum residual: {fit_error / plate_scale:.5g}")
        else:
            sys.stdout.write(
                f"* Final SIP degree: {deg}. "
                f"Maximum residual: {fit_error / plate_scale:.5g}"
            )

    return fit_poly_x, fit_poly_y, fit_error / plate_scale


def make_sampling_grid(npoints, bounding_box, crpix):
    step = np.subtract.reduce(bounding_box, axis=1) / (1.0 - npoints)
    crpix = np.asanyarray(crpix)[:, None, None]
    x, y = grid_from_bounding_box(bounding_box, step=step, center=False) - crpix
    return x.flatten(), y.flatten()


def _compute_distance_residual(undist_x, undist_y, fit_poly_x, fit_poly_y):
    """
    Compute the distance residuals and return the rms and maximum values.
    """
    dist = np.sqrt((undist_x - fit_poly_x) ** 2 + (undist_y - fit_poly_y) ** 2)
    return dist.max()


def reform_poly_coefficients(fit_poly_x, fit_poly_y):
    """
    The fit polynomials must be recombined to align with the SIP decomposition

    The result is the f(u,v) and g(u,v) polynomials, and the CD matrix.
    """
    # Extract values for CD matrix and recombining
    c11 = fit_poly_x.c1_0.value
    c12 = fit_poly_x.c0_1.value
    c21 = fit_poly_y.c1_0.value
    c22 = fit_poly_y.c0_1.value
    sip_poly_x = fit_poly_x.copy()
    sip_poly_y = fit_poly_y.copy()
    # Force low order coefficients to be 0 as defined in SIP
    sip_poly_x.c0_0 = 0
    sip_poly_y.c0_0 = 0
    sip_poly_x.c1_0 = 0
    sip_poly_x.c0_1 = 0
    sip_poly_y.c1_0 = 0
    sip_poly_y.c0_1 = 0

    cdmat = ((c11, c12), (c21, c22))
    invcdmat = npla.inv(np.array(cdmat))
    degree = fit_poly_x.degree
    # Now loop through all remaining coefficients
    for i in range(degree + 1):
        for j in range(degree + 1):
            if (i + j > 1) and (i + j < degree + 1):
                old_x = getattr(fit_poly_x, f"c{i}_{j}").value
                old_y = getattr(fit_poly_y, f"c{i}_{j}").value
                newcoeff = np.dot(invcdmat, np.array([[old_x], [old_y]]))
                setattr(sip_poly_x, f"c{i}_{j}", newcoeff[0, 0])
                setattr(sip_poly_y, f"c{i}_{j}", newcoeff[1, 0])

    return cdmat, sip_poly_x, sip_poly_y


def store_2D_coefficients(hdr, poly_model, coeff_prefix, keeplinear=False):
    """
    Write the polynomial model coefficients to the header.
    """
    mindeg = int(not keeplinear)
    degree = poly_model.degree
    for i in range(degree + 1):
        for j in range(degree + 1):
            if (i + j) > mindeg and (i + j < degree + 1):
                hdr[f"{coeff_prefix}_{i}_{j}"] = getattr(poly_model, f"c{i}_{j}").value


def fix_transform_inputs(transform, inputs):
    # This is a workaround to the bug in https://github.com/astropy/astropy/issues/11360
    # Once that bug is fixed, the code below can be replaced with fix_inputs
    if not inputs:
        return transform

    c = None
    mapping = []
    for k in range(transform.n_inputs):
        if k in inputs:
            mapping.append(0)
        else:
            # this assumes that n_inputs > 0 and that axis 0 always exist
            c = 0 if c is None else (c + 1)
            mapping.append(c)

    in_selector = Mapping(mapping, n_inputs=transform.n_inputs - len(inputs))

    input_fixer = Const1D(inputs[0]) if 0 in inputs else Identity(1)
    for k in range(1, transform.n_inputs):
        input_fixer &= Const1D(inputs[k]) if k in inputs else Identity(1)

    return in_selector | input_fixer | transform



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import io
from itertools import permutations, product

import asdf
import numpy as np
import pytest
from astropy import units as u

try:
    from asdf_astropy.testing.helpers import assert_model_roundtrip
except ImportError:
    from asdf_astropy.converters.transform.tests.test_transform import (
        assert_model_roundtrip,
    )

from gwcs import geometry

_INV_SQRT2 = 1.0 / np.sqrt(2.0)


def test_spherical_cartesian_inverse():
    t = geometry.SphericalToCartesian()
    assert isinstance(t.inverse, geometry.CartesianToSpherical)

    t = geometry.CartesianToSpherical()
    assert isinstance(t.inverse, geometry.SphericalToCartesian)


@pytest.mark.parametrize(
    ("testval", "unit", "wrap_at"),
    product(
        [
            (45.0, -90.0, (0.0, 0.0, -1.0)),
            (45.0, -45.0, (0.5, 0.5, -_INV_SQRT2)),
            (45, 0.0, (_INV_SQRT2, _INV_SQRT2, 0.0)),
            (45.0, 45, (0.5, 0.5, _INV_SQRT2)),
            (45.0, 90.0, (0.0, 0.0, 1.0)),
            (135.0, -90.0, (0.0, 0.0, -1.0)),
            (135.0, -45.0, (-0.5, 0.5, -_INV_SQRT2)),
            (135.0, 0.0, (-_INV_SQRT2, _INV_SQRT2, 0.0)),
            (135.0, 45.0, (-0.5, 0.5, _INV_SQRT2)),
            (135.0, 90.0, (0.0, 0.0, 1.0)),
            (225.0, -90.0, (0.0, 0.0, -1.0)),
            (225.0, -45.0, (-0.5, -0.5, -_INV_SQRT2)),
            (225.0, 0.0, (-_INV_SQRT2, -_INV_SQRT2, 0.0)),
            (225.0, 45.0, (-0.5, -0.5, _INV_SQRT2)),
            (225.0, 90.0, (0.0, 0.0, 1.0)),
            (315.0, -90.0, (0.0, 0.0, -1.0)),
            (315.0, -45.0, (0.5, -0.5, -_INV_SQRT2)),
            (315.0, 0.0, (_INV_SQRT2, -_INV_SQRT2, 0.0)),
            (315.0, 45.0, (0.5, -0.5, _INV_SQRT2)),
            (315.0, 90.0, (0.0, 0.0, 1.0)),
        ],
        [1, 1 * u.deg, 3600.0 * u.arcsec, np.pi / 180.0 * u.rad],
        [180, 360],
    ),
)
def test_spherical_to_cartesian(testval, unit, wrap_at):
    s2c = geometry.SphericalToCartesian(wrap_lon_at=wrap_at)
    ounit = 1 if unit == 1 else u.dimensionless_unscaled
    lon, lat, expected = testval

    if wrap_at == 180:
        lon = np.mod(lon - 180.0, 360.0) - 180.0

    xyz = s2c(lon * unit, lat * unit)
    if unit != 1:
        assert xyz[0].unit == u.dimensionless_unscaled
    assert u.allclose(xyz, u.Quantity(expected, ounit), atol=1e-15 * ounit)


@pytest.mark.parametrize(
    ("lon", "lat", "unit", "wrap_at"),
    list(
        product(
            [0, 45, 90, 135, 180, 225, 270, 315, 360],
            [-90, -89, -55, 0, 25, 89, 90],
            [1, 1 * u.deg, 3600.0 * u.arcsec, np.pi / 180.0 * u.rad],
            [180, 360],
        )
    ),
)
def test_spher2cart_roundrip(lon, lat, unit, wrap_at):
    s2c = geometry.SphericalToCartesian(wrap_lon_at=wrap_at)
    c2s = geometry.CartesianToSpherical(wrap_lon_at=wrap_at)
    ounit = 1 if unit == 1 else u.deg

    if wrap_at == 180:
        lon = np.mod(lon - 180.0, 360.0) - 180.0

    assert s2c.wrap_lon_at == wrap_at
    assert c2s.wrap_lon_at == wrap_at

    assert u.allclose(
        c2s(*s2c(lon * unit, lat * unit)),
        (lon * ounit, lat * ounit),
        atol=1e-15 * ounit,
    )


def test_cart2spher_at_pole(cart_to_spher):
    assert np.allclose(cart_to_spher(0, 0, 1), (0, 90), rtol=0, atol=1e-15)


@pytest.mark.parametrize(
    ("lonlat", "unit", "wrap_at"),
    list(
        product(
            [
                [[1], [-80]],
                [[325], [-89]],
                [[0, 1, 120, 180, 225, 325, 359], [-89, 0, 89, 10, -15, 45, -30]],
                [
                    np.array([0.0, 1, 120, 180, 225, 325, 359]),
                    np.array([-89, 0.0, 89, 10, -15, 45, -30]),
                ],
            ],
            [None, 1 * u.deg],
            [180, 360],
        )
    ),
)
def test_spher2cart_roundrip_arr(lonlat, unit, wrap_at):
    lon, lat = lonlat
    s2c = geometry.SphericalToCartesian(wrap_lon_at=wrap_at)
    c2s = geometry.CartesianToSpherical(wrap_lon_at=wrap_at)

    if wrap_at == 180:
        if isinstance(lon, np.ndarray):
            lon = np.mod(lon - 180.0, 360.0) - 180.0
        else:
            lon = [((deg - 180.0) % 360.0) - 180.0 for deg in lon]

    atol = 1e-15
    if unit is None:
        olon = lon
        olat = lat
    else:
        olon = lon * u.deg
        olat = lat * u.deg
        lon = lon * unit
        lat = lat * unit
        atol = atol * u.deg

    assert u.allclose(c2s(*s2c(lon, lat)), (olon, olat), atol=atol)


@pytest.mark.parametrize(("unit1", "unit2"), [(u.deg, 1), (1, u.deg)])
def test_spherical_to_cartesian_mixed_Q(spher_to_cart, unit1, unit2):
    with pytest.raises(TypeError) as arg_err:
        spher_to_cart(135.0 * unit1, 45.0 * unit2)
    assert (
        arg_err.value.args[0] == "All arguments must be of the same type "
        "(i.e., quantity or not)."
    )


@pytest.mark.parametrize(
    ("x", "y", "z"),
    sorted(
        set(
            tuple(permutations([1 * u.m, 1, 1]))
            + tuple(permutations([1 * u.m, 1 * u.m, 1]))
        ),
        key=str,
    ),
)
def test_cartesian_to_spherical_mixed_Q(cart_to_spher, x, y, z):
    with pytest.raises(TypeError) as arg_err:
        cart_to_spher(x, y, z)
    assert (
        arg_err.value.args[0] == "All arguments must be of the same type "
        "(i.e., quantity or not)."
    )


@pytest.mark.parametrize("wrap_at", ["1", 180.0, True, 180j, [180], -180, 0])
def test_c2s2c_wrong_wrap_type(spher_to_cart, cart_to_spher, wrap_at):
    err_msg = "'wrap_lon_at' must be an integer number: 180 or 360"
    with pytest.raises(ValueError, match=err_msg):
        geometry.SphericalToCartesian(wrap_lon_at=wrap_at)

    with pytest.raises(ValueError, match=err_msg):
        spher_to_cart.wrap_lon_at = wrap_at

    with pytest.raises(ValueError, match=err_msg):
        geometry.CartesianToSpherical(wrap_lon_at=wrap_at)

    with pytest.raises(ValueError, match=err_msg):
        cart_to_spher.wrap_lon_at = wrap_at


def test_cartesian_spherical_asdf(tmp_path):
    s2c0 = geometry.SphericalToCartesian(wrap_lon_at=360)
    c2s0 = geometry.CartesianToSpherical(wrap_lon_at=180)

    # asdf round-trip test:
    assert_model_roundtrip(c2s0, tmp_path)
    assert_model_roundtrip(s2c0, tmp_path)

    # create file object
    f = asdf.AsdfFile({"c2s": c2s0, "s2c": s2c0})

    # write to...
    buf = io.BytesIO()
    f.write_to(buf)

    # read back:
    buf.seek(0)
    f = asdf.open(buf)

    # retrieve transformations:
    c2s = f["c2s"]
    s2c = f["s2c"]

    pcoords = [
        (45.0, -90.0),
        (45.0, -45.0),
        (45, 0.0),
        (45.0, 45),
        (45.0, 90.0),
        (135.0, -90.0),
        (135.0, -45.0),
        (135.0, 0.0),
        (135.0, 45.0),
        (135.0, 90.0),
    ]

    ncoords = [
        (225.0, -90.0),
        (225.0, -45.0),
        (225.0, 0.0),
        (225.0, 45.0),
        (225.0, 90.0),
        (315.0, -90.0),
        (315.0, -45.0),
        (315.0, 0.0),
        (315.0, 45.0),
        (315.0, 90.0),
    ]

    for lon, lat in pcoords:
        xyz = s2c(lon, lat)
        assert xyz == s2c0(lon, lat)
        lon2, lat2 = c2s(*xyz)
        assert lon2, lat2 == c2s0(*xyz)
        assert np.allclose((lon, lat), (lon2, lat2))

    for lon, lat in ncoords:
        xyz = s2c(lon, lat)
        assert xyz == s2c0(lon, lat)
        lon2, lat2 = c2s(*xyz)
        lon3, lat3 = s2c.inverse(*xyz)
        assert lon2, lat2 == c2s0(*xyz)
        assert np.allclose((lon, lat), (lon2 + 360, lat2))
        assert np.allclose((lon, lat), (lon3, lat2))



"""
ASDF tags for geometry related models.

"""

from asdf_astropy.converters.transform.core import TransformConverterBase

__all__ = ["DirectionCosinesConverter", "SphericalCartesianConverter"]


class DirectionCosinesConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/direction_cosines-*",)
    types = (
        "gwcs.geometry.ToDirectionCosines",
        "gwcs.geometry.FromDirectionCosines",
    )

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.geometry import FromDirectionCosines, ToDirectionCosines

        transform_type = node["transform_type"]
        if transform_type == "to_direction_cosines":
            return ToDirectionCosines()
        if transform_type == "from_direction_cosines":
            return FromDirectionCosines()
        msg = f"Unknown model_type {transform_type}"
        raise TypeError(msg)

    def to_yaml_tree_transform(self, model, tag, ctx):
        from gwcs.geometry import FromDirectionCosines, ToDirectionCosines

        if isinstance(model, FromDirectionCosines):
            transform_type = "from_direction_cosines"
        elif isinstance(model, ToDirectionCosines):
            transform_type = "to_direction_cosines"
        else:
            msg = f"Model of type {model.__class__} is not supported."
            raise TypeError(msg)
        return {"transform_type": transform_type}


class SphericalCartesianConverter(TransformConverterBase):
    tags = ("tag:stsci.edu:gwcs/spherical_cartesian-*",)
    types = (
        "gwcs.geometry.SphericalToCartesian",
        "gwcs.geometry.CartesianToSpherical",
    )

    def from_yaml_tree_transform(self, node, tag, ctx):
        from gwcs.geometry import CartesianToSpherical, SphericalToCartesian

        transform_type = node["transform_type"]
        wrap_lon_at = node["wrap_lon_at"]
        if transform_type == "spherical_to_cartesian":
            return SphericalToCartesian(wrap_lon_at=wrap_lon_at)
        if transform_type == "cartesian_to_spherical":
            return CartesianToSpherical(wrap_lon_at=wrap_lon_at)
        msg = f"Unknown model_type {transform_type}"
        raise TypeError(msg)

    def to_yaml_tree_transform(self, model, tag, ctx):
        from gwcs.geometry import CartesianToSpherical, SphericalToCartesian

        if isinstance(model, SphericalToCartesian):
            transform_type = "spherical_to_cartesian"
        elif isinstance(model, CartesianToSpherical):
            transform_type = "cartesian_to_spherical"
        else:
            msg = f"Model of type {model.__class__} is not supported."
            raise TypeError(msg)

        return {"transform_type": transform_type, "wrap_lon_at": model.wrap_lon_at}
